import json

from django.test import override_settings

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.oauth2.rfc6749 import errors
from authlib.oauth2.rfc6749 import grants

from .models import Client
from .models import CodeGrantMixin
from .models import OAuth2Code
from .models import User
from .oauth2_server import TestCase


class AuthorizationCodeGrant(CodeGrantMixin, grants.AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        auth_code = OAuth2Code(
            code=code,
            client_id=request.client.client_id,
            redirect_uri=request.redirect_uri,
            response_type=request.response_type,
            scope=request.scope,
            user=request.user,
        )
        auth_code.save()


class AuthorizationCodeTest(TestCase):
    def create_server(self):
        server = super().create_server()
        server.register_grant(AuthorizationCodeGrant)
        return server

    def prepare_data(
        self, response_type="code", grant_type="authorization_code", scope=""
    ):
        user = User(username="foo")
        user.save()
        client = Client(
            user_id=user.pk,
            client_id="client",
            client_secret="secret",
            response_type=response_type,
            grant_type=grant_type,
            scope=scope,
            token_endpoint_auth_method="client_secret_basic",
            default_redirect_uri="https://a.b",
        )
        client.save()

    def test_get_consent_grant_client(self):
        server = self.create_server()
        url = "/authorize?response_type=code"
        request = self.factory.get(url)
        self.assertRaises(errors.InvalidClientError, server.get_consent_grant, request)

        url = "/authorize?response_type=code&client_id=client"
        request = self.factory.get(url)
        self.assertRaises(errors.InvalidClientError, server.get_consent_grant, request)

        self.prepare_data(response_type="")
        self.assertRaises(
            errors.UnauthorizedClientError, server.get_consent_grant, request
        )

        url = "/authorize?response_type=code&client_id=client&scope=profile&state=bar&redirect_uri=https%3A%2F%2Fa.b&response_type=code"
        request = self.factory.get(url)
        self.assertRaises(errors.InvalidRequestError, server.get_consent_grant, request)

    def test_get_consent_grant_redirect_uri(self):
        server = self.create_server()
        self.prepare_data()

        base_url = "/authorize?response_type=code&client_id=client"
        url = base_url + "&redirect_uri=https%3A%2F%2Fa.c"
        request = self.factory.get(url)
        self.assertRaises(errors.InvalidRequestError, server.get_consent_grant, request)

        url = base_url + "&redirect_uri=https%3A%2F%2Fa.b"
        request = self.factory.get(url)
        grant = server.get_consent_grant(request)
        self.assertIsInstance(grant, AuthorizationCodeGrant)

    def test_get_consent_grant_scope(self):
        server = self.create_server()
        server.scopes_supported = ["profile"]

        self.prepare_data()
        base_url = "/authorize?response_type=code&client_id=client"
        url = base_url + "&scope=invalid"
        request = self.factory.get(url)
        self.assertRaises(errors.InvalidScopeError, server.get_consent_grant, request)

    def test_create_authorization_response(self):
        server = self.create_server()
        self.prepare_data()
        data = {"response_type": "code", "client_id": "client"}
        request = self.factory.post("/authorize", data=data)
        server.get_consent_grant(request)

        resp = server.create_authorization_response(request)
        self.assertEqual(resp.status_code, 302)
        self.assertIn("error=access_denied", resp["Location"])

        grant_user = User.objects.get(username="foo")
        resp = server.create_authorization_response(request, grant_user=grant_user)
        self.assertEqual(resp.status_code, 302)
        self.assertIn("code=", resp["Location"])

    def test_create_token_response_invalid(self):
        server = self.create_server()
        self.prepare_data()

        # case: no auth
        request = self.factory.post(
            "/oauth/token", data={"grant_type": "authorization_code"}
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_client")

        auth_header = self.create_basic_auth("client", "secret")

        # case: no code
        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "authorization_code"},
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_request")

        # case: invalid code
        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "authorization_code", "code": "invalid"},
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_grant")

    def test_create_token_response_success(self):
        self.prepare_data()
        data = self.get_token_response()
        self.assertIn("access_token", data)
        self.assertNotIn("refresh_token", data)

    @override_settings(AUTHLIB_OAUTH2_PROVIDER={"refresh_token_generator": True})
    def test_create_token_response_with_refresh_token(self):
        self.prepare_data(grant_type="authorization_code\nrefresh_token")
        data = self.get_token_response()
        self.assertIn("access_token", data)
        self.assertIn("refresh_token", data)

    def get_token_response(self):
        server = self.create_server()
        data = {"response_type": "code", "client_id": "client"}
        request = self.factory.post("/authorize", data=data)
        grant_user = User.objects.get(username="foo")
        resp = server.create_authorization_response(request, grant_user=grant_user)
        self.assertEqual(resp.status_code, 302)

        params = dict(url_decode(urlparse.urlparse(resp["Location"]).query))
        code = params["code"]

        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "authorization_code", "code": code},
            HTTP_AUTHORIZATION=self.create_basic_auth("client", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content)
        return data



from flask import json

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)

from .models import AuthorizationCode
from .models import Client
from .models import CodeGrantMixin
from .models import User
from .models import db
from .models import save_authorization_code
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


class AuthorizationCodeTest(TestCase):
    LAZY_INIT = False

    def register_grant(self, server):
        server.register_grant(AuthorizationCodeGrant)

    def prepare_data(
        self,
        is_confidential=True,
        response_type="code",
        grant_type="authorization_code",
        token_endpoint_auth_method="client_secret_basic",
    ):
        server = create_authorization_server(self.app, self.LAZY_INIT)
        self.register_grant(server)
        self.server = server

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()

        if is_confidential:
            client_secret = "code-secret"
        else:
            client_secret = ""
        client = Client(
            user_id=user.id,
            client_id="code-client",
            client_secret=client_secret,
        )
        client.set_client_metadata(
            {
                "redirect_uris": ["https://a.b"],
                "scope": "profile address",
                "token_endpoint_auth_method": token_endpoint_auth_method,
                "response_types": [response_type],
                "grant_types": grant_type.splitlines(),
            }
        )
        self.authorize_url = "/oauth/authorize?response_type=code&client_id=code-client"
        db.session.add(client)
        db.session.commit()

    def test_get_authorize(self):
        self.prepare_data()
        rv = self.client.get(self.authorize_url)
        self.assertEqual(rv.data, b"ok")

    def test_invalid_client_id(self):
        self.prepare_data()
        url = "/oauth/authorize?response_type=code"
        rv = self.client.get(url)
        self.assertIn(b"invalid_client", rv.data)

        url = "/oauth/authorize?response_type=code&client_id=invalid"
        rv = self.client.get(url)
        self.assertIn(b"invalid_client", rv.data)

    def test_invalid_authorize(self):
        self.prepare_data()
        rv = self.client.post(self.authorize_url)
        self.assertIn("error=access_denied", rv.location)

        self.server.scopes_supported = ["profile"]
        rv = self.client.post(self.authorize_url + "&scope=invalid&state=foo")
        self.assertIn("error=invalid_scope", rv.location)
        self.assertIn("state=foo", rv.location)

    def test_unauthorized_client(self):
        self.prepare_data(True, "token")
        rv = self.client.get(self.authorize_url)
        self.assertIn(b"unauthorized_client", rv.data)

    def test_invalid_client(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": "invalid",
                "client_id": "invalid-id",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        headers = self.create_basic_header("code-client", "invalid-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": "invalid",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")
        self.assertEqual(resp["error_uri"], "https://a.b/e#invalid_client")

    def test_invalid_code(self):
        self.prepare_data()

        headers = self.create_basic_header("code-client", "code-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": "invalid",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_grant")

        code = AuthorizationCode(code="no-user", client_id="code-client", user_id=0)
        db.session.add(code)
        db.session.commit()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": "no-user",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_grant")

    def test_invalid_redirect_uri(self):
        self.prepare_data()
        uri = self.authorize_url + "&redirect_uri=https%3A%2F%2Fa.c"
        rv = self.client.post(uri, data={"user_id": "1"})
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

        uri = self.authorize_url + "&redirect_uri=https%3A%2F%2Fa.b"
        rv = self.client.post(uri, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        headers = self.create_basic_header("code-client", "code-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_grant")

    def test_invalid_grant_type(self):
        self.prepare_data(
            False, token_endpoint_auth_method="none", grant_type="invalid"
        )
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "client_id": "code-client",
                "code": "a",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unauthorized_client")

    def test_authorize_token_no_refresh_token(self):
        self.app.config.update({"OAUTH2_REFRESH_TOKEN_GENERATOR": True})
        self.prepare_data(False, token_endpoint_auth_method="none")

        rv = self.client.post(self.authorize_url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "client_id": "code-client",
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertNotIn("refresh_token", resp)

    def test_authorize_token_has_refresh_token(self):
        # generate refresh token
        self.app.config.update({"OAUTH2_REFRESH_TOKEN_GENERATOR": True})
        self.prepare_data(grant_type="authorization_code\nrefresh_token")
        url = self.authorize_url + "&state=bar"
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        self.assertEqual(params["state"], "bar")

        code = params["code"]
        headers = self.create_basic_header("code-client", "code-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("refresh_token", resp)

    def test_invalid_multiple_request_parameters(self):
        self.prepare_data()
        url = (
            self.authorize_url
            + "&scope=profile&state=bar&redirect_uri=https%3A%2F%2Fa.b&response_type=code"
        )
        rv = self.client.get(url)
        self.assertIn(b"invalid_request", rv.data)
        self.assertIn(b"Multiple+%22response_type%22+in+request.", rv.data)

    def test_client_secret_post(self):
        self.app.config.update({"OAUTH2_REFRESH_TOKEN_GENERATOR": True})
        self.prepare_data(
            grant_type="authorization_code\nrefresh_token",
            token_endpoint_auth_method="client_secret_post",
        )
        url = self.authorize_url + "&state=bar"
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        self.assertEqual(params["state"], "bar")

        code = params["code"]
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "client_id": "code-client",
                "client_secret": "code-secret",
                "code": code,
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("refresh_token", resp)

    def test_token_generator(self):
        m = "tests.flask.test_oauth2.oauth2_server:token_generator"
        self.app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
        self.prepare_data(False, token_endpoint_auth_method="none")

        rv = self.client.post(self.authorize_url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "client_id": "code-client",
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("c-authorization_code.1.", resp["access_token"])



from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.oauth2.rfc6749 import errors
from authlib.oauth2.rfc6749 import grants

from .models import Client
from .models import User
from .oauth2_server import TestCase


class ImplicitTest(TestCase):
    def create_server(self):
        server = super().create_server()
        server.register_grant(grants.ImplicitGrant)
        return server

    def prepare_data(self, response_type="token", scope=""):
        user = User(username="foo")
        user.save()
        client = Client(
            user_id=user.pk,
            client_id="client",
            response_type=response_type,
            scope=scope,
            token_endpoint_auth_method="none",
            default_redirect_uri="https://a.b",
        )
        client.save()

    def test_get_consent_grant_client(self):
        server = self.create_server()
        url = "/authorize?response_type=token"
        request = self.factory.get(url)
        self.assertRaises(errors.InvalidClientError, server.get_consent_grant, request)

        url = "/authorize?response_type=token&client_id=client"
        request = self.factory.get(url)
        self.assertRaises(errors.InvalidClientError, server.get_consent_grant, request)

        self.prepare_data(response_type="")
        self.assertRaises(
            errors.UnauthorizedClientError, server.get_consent_grant, request
        )

    def test_get_consent_grant_scope(self):
        server = self.create_server()
        server.scopes_supported = ["profile"]

        self.prepare_data()
        base_url = "/authorize?response_type=token&client_id=client"
        url = base_url + "&scope=invalid"
        request = self.factory.get(url)
        self.assertRaises(errors.InvalidScopeError, server.get_consent_grant, request)

    def test_create_authorization_response(self):
        server = self.create_server()
        self.prepare_data()
        data = {"response_type": "token", "client_id": "client"}
        request = self.factory.post("/authorize", data=data)
        server.get_consent_grant(request)

        resp = server.create_authorization_response(request)
        self.assertEqual(resp.status_code, 302)
        params = dict(url_decode(urlparse.urlparse(resp["Location"]).fragment))
        self.assertEqual(params["error"], "access_denied")

        grant_user = User.objects.get(username="foo")
        resp = server.create_authorization_response(request, grant_user=grant_user)
        self.assertEqual(resp.status_code, 302)
        params = dict(url_decode(urlparse.urlparse(resp["Location"]).fragment))
        self.assertIn("access_token", params)



import unittest

from authlib.oauth2.rfc8414 import AuthorizationServerMetadata
from authlib.oauth2.rfc8414 import get_well_known_url

WELL_KNOWN_URL = "/.well-known/oauth-authorization-server"


class WellKnownTest(unittest.TestCase):
    def test_no_suffix_issuer(self):
        self.assertEqual(get_well_known_url("https://authlib.org"), WELL_KNOWN_URL)
        self.assertEqual(get_well_known_url("https://authlib.org/"), WELL_KNOWN_URL)

    def test_with_suffix_issuer(self):
        self.assertEqual(
            get_well_known_url("https://authlib.org/issuer1"),
            WELL_KNOWN_URL + "/issuer1",
        )
        self.assertEqual(
            get_well_known_url("https://authlib.org/a/b/c"), WELL_KNOWN_URL + "/a/b/c"
        )

    def test_with_external(self):
        self.assertEqual(
            get_well_known_url("https://authlib.org", external=True),
            "https://authlib.org" + WELL_KNOWN_URL,
        )

    def test_with_changed_suffix(self):
        url = get_well_known_url("https://authlib.org", suffix="openid-configuration")
        self.assertEqual(url, "/.well-known/openid-configuration")
        url = get_well_known_url(
            "https://authlib.org", external=True, suffix="openid-configuration"
        )
        self.assertEqual(url, "https://authlib.org/.well-known/openid-configuration")


class AuthorizationServerMetadataTest(unittest.TestCase):
    def test_validate_issuer(self):
        #: missing
        metadata = AuthorizationServerMetadata({})
        with self.assertRaises(ValueError) as cm:
            metadata.validate()
        self.assertEqual('"issuer" is required', str(cm.exception))

        #: https
        metadata = AuthorizationServerMetadata({"issuer": "http://authlib.org/"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_issuer()
        self.assertIn("https", str(cm.exception))

        #: query
        metadata = AuthorizationServerMetadata({"issuer": "https://authlib.org/?a=b"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_issuer()
        self.assertIn("query", str(cm.exception))

        #: fragment
        metadata = AuthorizationServerMetadata({"issuer": "https://authlib.org/#a=b"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_issuer()
        self.assertIn("fragment", str(cm.exception))

        metadata = AuthorizationServerMetadata({"issuer": "https://authlib.org/"})
        metadata.validate_issuer()

    def test_validate_authorization_endpoint(self):
        # https
        metadata = AuthorizationServerMetadata(
            {"authorization_endpoint": "http://authlib.org/"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_authorization_endpoint()
        self.assertIn("https", str(cm.exception))

        # valid https
        metadata = AuthorizationServerMetadata(
            {"authorization_endpoint": "https://authlib.org/"}
        )
        metadata.validate_authorization_endpoint()

        # missing
        metadata = AuthorizationServerMetadata()
        with self.assertRaises(ValueError) as cm:
            metadata.validate_authorization_endpoint()
        self.assertIn("required", str(cm.exception))

        # valid missing
        metadata = AuthorizationServerMetadata({"grant_types_supported": ["password"]})
        metadata.validate_authorization_endpoint()

    def test_validate_token_endpoint(self):
        # implicit
        metadata = AuthorizationServerMetadata({"grant_types_supported": ["implicit"]})
        metadata.validate_token_endpoint()

        # missing
        metadata = AuthorizationServerMetadata()
        with self.assertRaises(ValueError) as cm:
            metadata.validate_token_endpoint()
        self.assertIn("required", str(cm.exception))

        # https
        metadata = AuthorizationServerMetadata(
            {"token_endpoint": "http://authlib.org/"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_token_endpoint()
        self.assertIn("https", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata(
            {"token_endpoint": "https://authlib.org/"}
        )
        metadata.validate_token_endpoint()

    def test_validate_jwks_uri(self):
        # can missing
        metadata = AuthorizationServerMetadata()
        metadata.validate_jwks_uri()

        metadata = AuthorizationServerMetadata(
            {"jwks_uri": "http://authlib.org/jwks.json"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_jwks_uri()
        self.assertIn("https", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {"jwks_uri": "https://authlib.org/jwks.json"}
        )
        metadata.validate_jwks_uri()

    def test_validate_registration_endpoint(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_registration_endpoint()

        metadata = AuthorizationServerMetadata(
            {"registration_endpoint": "http://authlib.org/"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_registration_endpoint()
        self.assertIn("https", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {"registration_endpoint": "https://authlib.org/"}
        )
        metadata.validate_registration_endpoint()

    def test_validate_scopes_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_scopes_supported()

        # not array
        metadata = AuthorizationServerMetadata({"scopes_supported": "foo"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_scopes_supported()
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata({"scopes_supported": ["foo"]})
        metadata.validate_scopes_supported()

    def test_validate_response_types_supported(self):
        # missing
        metadata = AuthorizationServerMetadata()
        with self.assertRaises(ValueError) as cm:
            metadata.validate_response_types_supported()
        self.assertIn("required", str(cm.exception))

        # not array
        metadata = AuthorizationServerMetadata({"response_types_supported": "code"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_response_types_supported()
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata({"response_types_supported": ["code"]})
        metadata.validate_response_types_supported()

    def test_validate_response_modes_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_response_modes_supported()

        # not array
        metadata = AuthorizationServerMetadata({"response_modes_supported": "query"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_response_modes_supported()
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata({"response_modes_supported": ["query"]})
        metadata.validate_response_modes_supported()

    def test_validate_grant_types_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_grant_types_supported()

        # not array
        metadata = AuthorizationServerMetadata({"grant_types_supported": "password"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_grant_types_supported()
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata({"grant_types_supported": ["password"]})
        metadata.validate_grant_types_supported()

    def test_validate_token_endpoint_auth_methods_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_token_endpoint_auth_methods_supported()

        # not array
        metadata = AuthorizationServerMetadata(
            {"token_endpoint_auth_methods_supported": "client_secret_basic"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_token_endpoint_auth_methods_supported()
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata(
            {"token_endpoint_auth_methods_supported": ["client_secret_basic"]}
        )
        metadata.validate_token_endpoint_auth_methods_supported()

    def test_validate_token_endpoint_auth_signing_alg_values_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_token_endpoint_auth_signing_alg_values_supported()

        metadata = AuthorizationServerMetadata(
            {"token_endpoint_auth_methods_supported": ["client_secret_jwt"]}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_token_endpoint_auth_signing_alg_values_supported()
        self.assertIn("required", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {"token_endpoint_auth_signing_alg_values_supported": "RS256"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_token_endpoint_auth_signing_alg_values_supported()
        self.assertIn("JSON array", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {
                "token_endpoint_auth_methods_supported": ["client_secret_jwt"],
                "token_endpoint_auth_signing_alg_values_supported": ["RS256", "none"],
            }
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_token_endpoint_auth_signing_alg_values_supported()
        self.assertIn("none", str(cm.exception))

    def test_validate_service_documentation(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_service_documentation()

        metadata = AuthorizationServerMetadata({"service_documentation": "invalid"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_service_documentation()
        self.assertIn("MUST be a URL", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {"service_documentation": "https://authlib.org/"}
        )
        metadata.validate_service_documentation()

    def test_validate_ui_locales_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_ui_locales_supported()

        # not array
        metadata = AuthorizationServerMetadata({"ui_locales_supported": "en"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_ui_locales_supported()
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata({"ui_locales_supported": ["en"]})
        metadata.validate_ui_locales_supported()

    def test_validate_op_policy_uri(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_op_policy_uri()

        metadata = AuthorizationServerMetadata({"op_policy_uri": "invalid"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_op_policy_uri()
        self.assertIn("MUST be a URL", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {"op_policy_uri": "https://authlib.org/"}
        )
        metadata.validate_op_policy_uri()

    def test_validate_op_tos_uri(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_op_tos_uri()

        metadata = AuthorizationServerMetadata({"op_tos_uri": "invalid"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_op_tos_uri()
        self.assertIn("MUST be a URL", str(cm.exception))

        metadata = AuthorizationServerMetadata({"op_tos_uri": "https://authlib.org/"})
        metadata.validate_op_tos_uri()

    def test_validate_revocation_endpoint(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_revocation_endpoint()

        # https
        metadata = AuthorizationServerMetadata(
            {"revocation_endpoint": "http://authlib.org/"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_revocation_endpoint()
        self.assertIn("https", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata(
            {"revocation_endpoint": "https://authlib.org/"}
        )
        metadata.validate_revocation_endpoint()

    def test_validate_revocation_endpoint_auth_methods_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_revocation_endpoint_auth_methods_supported()

        # not array
        metadata = AuthorizationServerMetadata(
            {"revocation_endpoint_auth_methods_supported": "client_secret_basic"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_revocation_endpoint_auth_methods_supported()
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata(
            {"revocation_endpoint_auth_methods_supported": ["client_secret_basic"]}
        )
        metadata.validate_revocation_endpoint_auth_methods_supported()

    def test_validate_revocation_endpoint_auth_signing_alg_values_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_revocation_endpoint_auth_signing_alg_values_supported()

        metadata = AuthorizationServerMetadata(
            {"revocation_endpoint_auth_methods_supported": ["client_secret_jwt"]}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_revocation_endpoint_auth_signing_alg_values_supported()
        self.assertIn("required", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {"revocation_endpoint_auth_signing_alg_values_supported": "RS256"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_revocation_endpoint_auth_signing_alg_values_supported()
        self.assertIn("JSON array", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {
                "revocation_endpoint_auth_methods_supported": ["client_secret_jwt"],
                "revocation_endpoint_auth_signing_alg_values_supported": [
                    "RS256",
                    "none",
                ],
            }
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_revocation_endpoint_auth_signing_alg_values_supported()
        self.assertIn("none", str(cm.exception))

    def test_validate_introspection_endpoint(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_introspection_endpoint()

        # https
        metadata = AuthorizationServerMetadata(
            {"introspection_endpoint": "http://authlib.org/"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_introspection_endpoint()
        self.assertIn("https", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata(
            {"introspection_endpoint": "https://authlib.org/"}
        )
        metadata.validate_introspection_endpoint()

    def test_validate_introspection_endpoint_auth_methods_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_introspection_endpoint_auth_methods_supported()

        # not array
        metadata = AuthorizationServerMetadata(
            {"introspection_endpoint_auth_methods_supported": "client_secret_basic"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_introspection_endpoint_auth_methods_supported()
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata(
            {"introspection_endpoint_auth_methods_supported": ["client_secret_basic"]}
        )
        metadata.validate_introspection_endpoint_auth_methods_supported()

    def test_validate_introspection_endpoint_auth_signing_alg_values_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_introspection_endpoint_auth_signing_alg_values_supported()

        metadata = AuthorizationServerMetadata(
            {"introspection_endpoint_auth_methods_supported": ["client_secret_jwt"]}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_introspection_endpoint_auth_signing_alg_values_supported()
        self.assertIn("required", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {"introspection_endpoint_auth_signing_alg_values_supported": "RS256"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_introspection_endpoint_auth_signing_alg_values_supported()
        self.assertIn("JSON array", str(cm.exception))

        metadata = AuthorizationServerMetadata(
            {
                "introspection_endpoint_auth_methods_supported": ["client_secret_jwt"],
                "introspection_endpoint_auth_signing_alg_values_supported": [
                    "RS256",
                    "none",
                ],
            }
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_introspection_endpoint_auth_signing_alg_values_supported()
        self.assertIn("none", str(cm.exception))

    def test_validate_code_challenge_methods_supported(self):
        metadata = AuthorizationServerMetadata()
        metadata.validate_code_challenge_methods_supported()

        # not array
        metadata = AuthorizationServerMetadata(
            {"code_challenge_methods_supported": "S256"}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_code_challenge_methods_supported()
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = AuthorizationServerMetadata(
            {"code_challenge_methods_supported": ["S256"]}
        )
        metadata.validate_code_challenge_methods_supported()



import json

from authlib.oauth2.rfc6749.grants import (
    ResourceOwnerPasswordCredentialsGrant as _PasswordGrant,
)

from .models import Client
from .models import User
from .oauth2_server import TestCase


class PasswordGrant(_PasswordGrant):
    def authenticate_user(self, username, password):
        try:
            user = User.objects.get(username=username)
            if user.check_password(password):
                return user
        except User.DoesNotExist:
            return None


class PasswordTest(TestCase):
    def create_server(self):
        server = super().create_server()
        server.register_grant(PasswordGrant)
        return server

    def prepare_data(self, grant_type="password", scope=""):
        user = User(username="foo")
        user.set_password("ok")
        user.save()
        client = Client(
            user_id=user.pk,
            client_id="client",
            client_secret="secret",
            scope=scope,
            grant_type=grant_type,
            token_endpoint_auth_method="client_secret_basic",
            default_redirect_uri="https://a.b",
        )
        client.save()

    def test_invalid_client(self):
        server = self.create_server()
        self.prepare_data()
        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "password", "username": "foo", "password": "ok"},
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_client")

        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "password", "username": "foo", "password": "ok"},
            HTTP_AUTHORIZATION=self.create_basic_auth("invalid", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_client")

    def test_invalid_scope(self):
        server = self.create_server()
        server.scopes_supported = ["profile"]
        self.prepare_data()
        request = self.factory.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
                "scope": "invalid",
            },
            HTTP_AUTHORIZATION=self.create_basic_auth("client", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_scope")

    def test_invalid_request(self):
        server = self.create_server()
        self.prepare_data()
        auth_header = self.create_basic_auth("client", "secret")

        # case 1
        request = self.factory.get(
            "/oauth/token?grant_type=password",
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "unsupported_grant_type")

        # case 2
        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "password"},
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_request")

        # case 3
        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "password", "username": "foo"},
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_request")

        # case 4
        request = self.factory.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "wrong",
            },
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_request")

    def test_unauthorized_client(self):
        server = self.create_server()
        self.prepare_data(grant_type="invalid")
        request = self.factory.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
            },
            HTTP_AUTHORIZATION=self.create_basic_auth("client", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "unauthorized_client")

    def test_authorize_token(self):
        server = self.create_server()
        self.prepare_data()
        request = self.factory.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
            },
            HTTP_AUTHORIZATION=self.create_basic_auth("client", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content)
        self.assertIn("access_token", data)



from starlette.datastructures import URL
from starlette.responses import RedirectResponse

from ..base_client import BaseApp
from ..base_client import OAuthError
from ..base_client.async_app import AsyncOAuth1Mixin
from ..base_client.async_app import AsyncOAuth2Mixin
from ..base_client.async_openid import AsyncOpenIDMixin
from ..httpx_client import AsyncOAuth1Client
from ..httpx_client import AsyncOAuth2Client


class StarletteAppMixin:
    async def save_authorize_data(self, request, **kwargs):
        state = kwargs.pop("state", None)
        if state:
            if self.framework.cache:
                session = None
            else:
                session = request.session
            await self.framework.set_state_data(session, state, kwargs)
        else:
            raise RuntimeError("Missing state value")

    async def authorize_redirect(self, request, redirect_uri=None, **kwargs):
        """Create a HTTP Redirect for Authorization Endpoint.

        :param request: HTTP request instance from Starlette view.
        :param redirect_uri: Callback or redirect URI for authorization.
        :param kwargs: Extra parameters to include.
        :return: A HTTP redirect response.
        """
        # Handle Starlette >= 0.26.0 where redirect_uri may now be a URL and not a string
        if redirect_uri and isinstance(redirect_uri, URL):
            redirect_uri = str(redirect_uri)
        rv = await self.create_authorization_url(redirect_uri, **kwargs)
        await self.save_authorize_data(request, redirect_uri=redirect_uri, **rv)
        return RedirectResponse(rv["url"], status_code=302)


class StarletteOAuth1App(StarletteAppMixin, AsyncOAuth1Mixin, BaseApp):
    client_cls = AsyncOAuth1Client

    async def authorize_access_token(self, request, **kwargs):
        params = dict(request.query_params)
        state = params.get("oauth_token")
        if not state:
            raise OAuthError(description='Missing "oauth_token" parameter')

        data = await self.framework.get_state_data(request.session, state)
        if not data:
            raise OAuthError(description='Missing "request_token" in temporary data')

        params["request_token"] = data["request_token"]
        params.update(kwargs)
        await self.framework.clear_state_data(request.session, state)
        return await self.fetch_access_token(**params)


class StarletteOAuth2App(
    StarletteAppMixin, AsyncOAuth2Mixin, AsyncOpenIDMixin, BaseApp
):
    client_cls = AsyncOAuth2Client

    async def authorize_access_token(self, request, **kwargs):
        error = request.query_params.get("error")
        if error:
            description = request.query_params.get("error_description")
            raise OAuthError(error=error, description=description)

        params = {
            "code": request.query_params.get("code"),
            "state": request.query_params.get("state"),
        }

        if self.framework.cache:
            session = None
        else:
            session = request.session

        claims_options = kwargs.pop("claims_options", None)
        state_data = await self.framework.get_state_data(session, params.get("state"))
        await self.framework.clear_state_data(session, params.get("state"))
        params = self._format_state_params(state_data, params)
        token = await self.fetch_access_token(**params, **kwargs)

        if "id_token" in token and "nonce" in state_data:
            userinfo = await self.parse_id_token(
                token, nonce=state_data["nonce"], claims_options=claims_options
            )
            token["userinfo"] = userinfo
        return token



from flask import json

from authlib.oauth2.rfc7523 import JWTBearerGrant as _JWTBearerGrant
from authlib.oauth2.rfc7523 import JWTBearerTokenGenerator
from tests.util import read_file_path

from .models import Client
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class JWTBearerGrant(_JWTBearerGrant):
    def resolve_issuer_client(self, issuer):
        return Client.query.filter_by(client_id=issuer).first()

    def resolve_client_key(self, client, headers, payload):
        keys = {"1": "foo", "2": "bar"}
        return keys[headers["kid"]]

    def authenticate_user(self, subject):
        return None

    def has_granted_permission(self, client, user):
        return True


class JWTBearerGrantTest(TestCase):
    def prepare_data(self, grant_type=None, token_generator=None):
        server = create_authorization_server(self.app)
        server.register_grant(JWTBearerGrant)

        if token_generator:
            server.register_token_generator(JWTBearerGrant.GRANT_TYPE, token_generator)

        if grant_type is None:
            grant_type = JWTBearerGrant.GRANT_TYPE

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="jwt-client",
            client_secret="jwt-secret",
        )
        client.set_client_metadata(
            {
                "scope": "profile",
                "redirect_uris": ["http://localhost/authorized"],
                "grant_types": [grant_type],
            }
        )
        db.session.add(client)
        db.session.commit()

    def test_missing_assertion(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token", data={"grant_type": JWTBearerGrant.GRANT_TYPE}
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")
        self.assertIn("assertion", resp["error_description"])

    def test_invalid_assertion(self):
        self.prepare_data()
        assertion = JWTBearerGrant.sign(
            "foo",
            issuer="jwt-client",
            audience="https://i.b/token",
            subject="none",
            header={"alg": "HS256", "kid": "1"},
        )
        rv = self.client.post(
            "/oauth/token",
            data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_grant")

    def test_authorize_token(self):
        self.prepare_data()
        assertion = JWTBearerGrant.sign(
            "foo",
            issuer="jwt-client",
            audience="https://i.b/token",
            subject=None,
            header={"alg": "HS256", "kid": "1"},
        )
        rv = self.client.post(
            "/oauth/token",
            data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_unauthorized_client(self):
        self.prepare_data("password")
        assertion = JWTBearerGrant.sign(
            "bar",
            issuer="jwt-client",
            audience="https://i.b/token",
            subject=None,
            header={"alg": "HS256", "kid": "2"},
        )
        rv = self.client.post(
            "/oauth/token",
            data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unauthorized_client")

    def test_token_generator(self):
        m = "tests.flask.test_oauth2.oauth2_server:token_generator"
        self.app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
        self.prepare_data()
        assertion = JWTBearerGrant.sign(
            "foo",
            issuer="jwt-client",
            audience="https://i.b/token",
            subject=None,
            header={"alg": "HS256", "kid": "1"},
        )
        rv = self.client.post(
            "/oauth/token",
            data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("j-", resp["access_token"])

    def test_jwt_bearer_token_generator(self):
        private_key = read_file_path("jwks_private.json")
        self.prepare_data(token_generator=JWTBearerTokenGenerator(private_key))
        assertion = JWTBearerGrant.sign(
            "foo",
            issuer="jwt-client",
            audience="https://i.b/token",
            subject=None,
            header={"alg": "HS256", "kid": "1"},
        )
        rv = self.client.post(
            "/oauth/token",
            data={"grant_type": JWTBearerGrant.GRANT_TYPE, "assertion": assertion},
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertEqual(resp["access_token"].count("."), 2)



import time

import pytest
from flask import json
from flask import jsonify

from authlib.common.security import generate_token
from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.integrations.flask_oauth2 import current_token
from authlib.jose import jwt
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oauth2.rfc7009 import RevocationEndpoint
from authlib.oauth2.rfc7662 import IntrospectionEndpoint
from authlib.oauth2.rfc9068 import JWTBearerTokenGenerator
from authlib.oauth2.rfc9068 import JWTBearerTokenValidator
from authlib.oauth2.rfc9068 import JWTIntrospectionEndpoint
from authlib.oauth2.rfc9068 import JWTRevocationEndpoint
from tests.util import read_file_path

from .models import Client
from .models import CodeGrantMixin
from .models import Token
from .models import User
from .models import db
from .models import save_authorization_code
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


def create_token_validator(issuer, resource_server, jwks):
    class MyJWTBearerTokenValidator(JWTBearerTokenValidator):
        def get_jwks(self):
            return jwks

    validator = MyJWTBearerTokenValidator(
        issuer=issuer, resource_server=resource_server
    )
    return validator


def create_resource_protector(app, validator):
    require_oauth = ResourceProtector()
    require_oauth.register_token_validator(validator)

    @app.route("/protected")
    @require_oauth()
    def protected():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    @app.route("/protected-by-scope")
    @require_oauth("profile")
    def protected_by_scope():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    @app.route("/protected-by-groups")
    @require_oauth(groups=["admins"])
    def protected_by_groups():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    @app.route("/protected-by-roles")
    @require_oauth(roles=["student"])
    def protected_by_roles():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    @app.route("/protected-by-entitlements")
    @require_oauth(entitlements=["captain"])
    def protected_by_entitlements():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    return require_oauth


def create_token_generator(authorization_server, issuer, jwks):
    class MyJWTBearerTokenGenerator(JWTBearerTokenGenerator):
        def get_jwks(self):
            return jwks

    token_generator = MyJWTBearerTokenGenerator(issuer=issuer)
    authorization_server.register_token_generator("default", token_generator)
    return token_generator


def create_introspection_endpoint(app, authorization_server, issuer, jwks):
    class MyJWTIntrospectionEndpoint(JWTIntrospectionEndpoint):
        def get_jwks(self):
            return jwks

        def check_permission(self, token, client, request):
            return client.client_id == "client-id"

    endpoint = MyJWTIntrospectionEndpoint(issuer=issuer)
    authorization_server.register_endpoint(endpoint)

    @app.route("/oauth/introspect", methods=["POST"])
    def introspect_token():
        return authorization_server.create_endpoint_response(
            MyJWTIntrospectionEndpoint.ENDPOINT_NAME
        )

    return endpoint


def create_revocation_endpoint(app, authorization_server, issuer, jwks):
    class MyJWTRevocationEndpoint(JWTRevocationEndpoint):
        def get_jwks(self):
            return jwks

    endpoint = MyJWTRevocationEndpoint(issuer=issuer)
    authorization_server.register_endpoint(endpoint)

    @app.route("/oauth/revoke", methods=["POST"])
    def revoke_token():
        return authorization_server.create_endpoint_response(
            MyJWTRevocationEndpoint.ENDPOINT_NAME
        )

    return endpoint


def create_user():
    user = User(username="foo")
    db.session.add(user)
    db.session.commit()
    return user


def create_oauth_client(client_id, user):
    oauth_client = Client(
        user_id=user.id,
        client_id=client_id,
        client_secret=client_id,
    )
    oauth_client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["http://localhost/authorized"],
            "response_types": ["code"],
            "token_endpoint_auth_method": "client_secret_post",
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(oauth_client)
    db.session.commit()
    return oauth_client


def create_access_token_claims(client, user, issuer, **kwargs):
    now = int(time.time())
    expires_in = now + 3600
    auth_time = now - 60

    return {
        "iss": kwargs.get("issuer", issuer),
        "exp": kwargs.get("exp", expires_in),
        "aud": kwargs.get("aud", client.client_id),
        "sub": kwargs.get("sub", user.get_user_id()),
        "client_id": kwargs.get("client_id", client.client_id),
        "iat": kwargs.get("iat", now),
        "jti": kwargs.get("jti", generate_token(16)),
        "auth_time": kwargs.get("auth_time", auth_time),
        "scope": kwargs.get("scope", client.scope),
        "groups": kwargs.get("groups", ["admins"]),
        "roles": kwargs.get("groups", ["student"]),
        "entitlements": kwargs.get("groups", ["captain"]),
    }


def create_access_token(claims, jwks, alg="RS256", typ="at+jwt"):
    header = {"alg": alg, "typ": typ}
    access_token = jwt.encode(
        header,
        claims,
        key=jwks,
        check=False,
    )
    return access_token.decode()


def create_token(access_token):
    token = Token(
        user_id=1,
        client_id="resource-server",
        token_type="bearer",
        access_token=access_token,
        scope="profile",
        expires_in=3600,
    )
    db.session.add(token)
    db.session.commit()
    return token


class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


class JWTAccessTokenGenerationTest(TestCase):
    def setUp(self):
        super().setUp()
        self.issuer = "https://authlib.org/"
        self.jwks = read_file_path("jwks_private.json")
        self.authorization_server = create_authorization_server(self.app)
        self.authorization_server.register_grant(AuthorizationCodeGrant)
        self.token_generator = create_token_generator(
            self.authorization_server, self.issuer, self.jwks
        )
        self.user = create_user()
        self.oauth_client = create_oauth_client("client-id", self.user)

    def test_generate_jwt_access_token(self):
        res = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": self.oauth_client.response_types[0],
                "client_id": self.oauth_client.client_id,
                "redirect_uri": self.oauth_client.redirect_uris[0],
                "scope": self.oauth_client.scope,
                "user_id": self.user.id,
            },
        )

        params = dict(url_decode(urlparse.urlparse(res.location).query))
        code = params["code"]
        res = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "client_id": self.oauth_client.client_id,
                "client_secret": self.oauth_client.client_secret,
                "scope": " ".join(self.oauth_client.scope),
                "redirect_uri": self.oauth_client.redirect_uris[0],
            },
        )

        access_token = res.json["access_token"]
        claims = jwt.decode(access_token, self.jwks)

        assert claims["iss"] == self.issuer
        assert claims["sub"] == self.user.id
        assert claims["scope"] == self.oauth_client.scope
        assert claims["client_id"] == self.oauth_client.client_id

        # This specification registers the 'application/at+jwt' media type, which can
        # be used to indicate that the content is a JWT access token. JWT access tokens
        # MUST include this media type in the 'typ' header parameter to explicitly
        # declare that the JWT represents an access token complying with this profile.
        # Per the definition of 'typ' in Section 4.1.9 of [RFC7515], it is RECOMMENDED
        # that the 'application/' prefix be omitted. Therefore, the 'typ' value used
        # SHOULD be 'at+jwt'.

        assert claims.header["typ"] == "at+jwt"

    def test_generate_jwt_access_token_extra_claims(self):
        """Authorization servers MAY return arbitrary attributes not defined in any
        existing specification, as long as the corresponding claim names are collision
        resistant or the access tokens are meant to be used only within a private
        subsystem. Please refer to Sections 4.2 and 4.3 of [RFC7519] for details.
        """

        def get_extra_claims(client, grant_type, user, scope):
            return {"username": user.username}

        self.token_generator.get_extra_claims = get_extra_claims

        res = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": self.oauth_client.response_types[0],
                "client_id": self.oauth_client.client_id,
                "redirect_uri": self.oauth_client.redirect_uris[0],
                "scope": self.oauth_client.scope,
                "user_id": self.user.id,
            },
        )

        params = dict(url_decode(urlparse.urlparse(res.location).query))
        code = params["code"]
        res = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "client_id": self.oauth_client.client_id,
                "client_secret": self.oauth_client.client_secret,
                "scope": " ".join(self.oauth_client.scope),
                "redirect_uri": self.oauth_client.redirect_uris[0],
            },
        )

        access_token = res.json["access_token"]
        claims = jwt.decode(access_token, self.jwks)
        assert claims["username"] == self.user.username

    @pytest.mark.skip
    def test_generate_jwt_access_token_no_user(self):
        res = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": self.oauth_client.response_types[0],
                "client_id": self.oauth_client.client_id,
                "redirect_uri": self.oauth_client.redirect_uris[0],
                "scope": self.oauth_client.scope,
                #'user_id': self.user.id,
            },
        )

        params = dict(url_decode(urlparse.urlparse(res.location).query))
        code = params["code"]
        res = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "client_id": self.oauth_client.client_id,
                "client_secret": self.oauth_client.client_secret,
                "scope": " ".join(self.oauth_client.scope),
                "redirect_uri": self.oauth_client.redirect_uris[0],
            },
        )

        access_token = res.json["access_token"]
        claims = jwt.decode(access_token, self.jwks)

        assert claims["sub"] == self.oauth_client.client_id

    def test_optional_fields(self):
        self.token_generator.get_auth_time = lambda *args: 1234
        self.token_generator.get_amr = lambda *args: "amr"
        self.token_generator.get_acr = lambda *args: "acr"

        res = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": self.oauth_client.response_types[0],
                "client_id": self.oauth_client.client_id,
                "redirect_uri": self.oauth_client.redirect_uris[0],
                "scope": self.oauth_client.scope,
                "user_id": self.user.id,
            },
        )

        params = dict(url_decode(urlparse.urlparse(res.location).query))
        code = params["code"]
        res = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "client_id": self.oauth_client.client_id,
                "client_secret": self.oauth_client.client_secret,
                "scope": " ".join(self.oauth_client.scope),
                "redirect_uri": self.oauth_client.redirect_uris[0],
            },
        )

        access_token = res.json["access_token"]
        claims = jwt.decode(access_token, self.jwks)

        assert claims["auth_time"] == 1234
        assert claims["amr"] == "amr"
        assert claims["acr"] == "acr"


class JWTAccessTokenResourceServerTest(TestCase):
    def setUp(self):
        super().setUp()
        self.issuer = "https://authorization-server.example.org/"
        self.resource_server = "resource-server-id"
        self.jwks = read_file_path("jwks_private.json")
        self.token_validator = create_token_validator(
            self.issuer, self.resource_server, self.jwks
        )
        self.resource_protector = create_resource_protector(
            self.app, self.token_validator
        )
        self.user = create_user()
        self.oauth_client = create_oauth_client(self.resource_server, self.user)
        self.claims = create_access_token_claims(
            self.oauth_client, self.user, self.issuer
        )
        self.access_token = create_access_token(self.claims, self.jwks)
        self.token = create_token(self.access_token)

    def test_access_resource(self):
        headers = {"Authorization": f"Bearer {self.access_token}"}

        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["username"], "foo")

    def test_missing_authorization(self):
        rv = self.client.get("/protected")
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "missing_authorization")

    def test_unsupported_token_type(self):
        headers = {"Authorization": "invalid token"}
        rv = self.client.get("/protected", headers=headers)
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unsupported_token_type")

    def test_invalid_token(self):
        headers = {"Authorization": "Bearer invalid"}
        rv = self.client.get("/protected", headers=headers)
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")

    def test_typ(self):
        """The resource server MUST verify that the 'typ' header value is 'at+jwt' or
        'application/at+jwt' and reject tokens carrying any other value.
        """
        access_token = create_access_token(self.claims, self.jwks, typ="at+jwt")

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["username"], "foo")

        access_token = create_access_token(
            self.claims, self.jwks, typ="application/at+jwt"
        )

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["username"], "foo")

        access_token = create_access_token(self.claims, self.jwks, typ="invalid")

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")

    def test_missing_required_claims(self):
        required_claims = ["iss", "exp", "aud", "sub", "client_id", "iat", "jti"]
        for claim in required_claims:
            claims = create_access_token_claims(
                self.oauth_client, self.user, self.issuer
            )
            del claims[claim]
            access_token = create_access_token(claims, self.jwks)

            headers = {"Authorization": f"Bearer {access_token}"}
            rv = self.client.get("/protected", headers=headers)
            resp = json.loads(rv.data)
            self.assertEqual(resp["error"], "invalid_token")

    def test_invalid_iss(self):
        """The issuer identifier for the authorization server (which is typically obtained
        during discovery) MUST exactly match the value of the 'iss' claim.
        """
        self.claims["iss"] = "invalid-issuer"
        access_token = create_access_token(self.claims, self.jwks)

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")

    def test_invalid_aud(self):
        """The resource server MUST validate that the 'aud' claim contains a resource
        indicator value corresponding to an identifier the resource server expects for
        itself. The JWT access token MUST be rejected if 'aud' does not contain a
        resource indicator of the current resource server as a valid audience.
        """
        self.claims["aud"] = "invalid-resource-indicator"
        access_token = create_access_token(self.claims, self.jwks)

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")

    def test_invalid_exp(self):
        """The current time MUST be before the time represented by the 'exp' claim.
        Implementers MAY provide for some small leeway, usually no more than a few
        minutes, to account for clock skew.
        """
        self.claims["exp"] = time.time() - 1
        access_token = create_access_token(self.claims, self.jwks)

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")

    def test_scope_restriction(self):
        """If an authorization request includes a scope parameter, the corresponding
        issued JWT access token SHOULD include a 'scope' claim as defined in Section
        4.2 of [RFC8693]. All the individual scope strings in the 'scope' claim MUST
        have meaning for the resources indicated in the 'aud' claim. See Section 5 for
        more considerations about the relationship between scope strings and resources
        indicated by the 'aud' claim.
        """
        self.claims["scope"] = ["invalid-scope"]
        access_token = create_access_token(self.claims, self.jwks)

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["username"], "foo")

        rv = self.client.get("/protected-by-scope", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "insufficient_scope")

    def test_entitlements_restriction(self):
        """Many authorization servers embed authorization attributes that go beyond the
        delegated scenarios described by [RFC7519] in the access tokens they issue.
        Typical examples include resource owner memberships in roles and groups that
        are relevant to the resource being accessed, entitlements assigned to the
        resource owner for the targeted resource that the authorization server knows
        about, and so on. An authorization server wanting to include such attributes
        in a JWT access token SHOULD use the 'groups', 'roles', and 'entitlements'
        attributes of the 'User' resource schema defined by Section 4.1.2 of
        [RFC7643]) as claim types.
        """
        for claim in ["groups", "roles", "entitlements"]:
            claims = create_access_token_claims(
                self.oauth_client, self.user, self.issuer
            )
            claims[claim] = ["invalid"]
            access_token = create_access_token(claims, self.jwks)

            headers = {"Authorization": f"Bearer {access_token}"}
            rv = self.client.get("/protected", headers=headers)
            resp = json.loads(rv.data)
            self.assertEqual(resp["username"], "foo")

            rv = self.client.get(f"/protected-by-{claim}", headers=headers)
            resp = json.loads(rv.data)
            self.assertEqual(resp["error"], "invalid_token")

    def test_extra_attributes(self):
        """Authorization servers MAY return arbitrary attributes not defined in any
        existing specification, as long as the corresponding claim names are collision
        resistant or the access tokens are meant to be used only within a private
        subsystem. Please refer to Sections 4.2 and 4.3 of [RFC7519] for details.
        """
        self.claims["email"] = "user@example.org"
        access_token = create_access_token(self.claims, self.jwks)

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["token"]["email"], "user@example.org")

    def test_invalid_auth_time(self):
        self.claims["auth_time"] = "invalid-auth-time"
        access_token = create_access_token(self.claims, self.jwks)

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")

    def test_invalid_amr(self):
        self.claims["amr"] = "invalid-amr"
        access_token = create_access_token(self.claims, self.jwks)

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = self.client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")


class JWTAccessTokenIntrospectionTest(TestCase):
    def setUp(self):
        super().setUp()
        self.issuer = "https://authlib.org/"
        self.resource_server = "resource-server-id"
        self.jwks = read_file_path("jwks_private.json")
        self.authorization_server = create_authorization_server(self.app)
        self.authorization_server.register_grant(AuthorizationCodeGrant)
        self.introspection_endpoint = create_introspection_endpoint(
            self.app, self.authorization_server, self.issuer, self.jwks
        )
        self.user = create_user()
        self.oauth_client = create_oauth_client("client-id", self.user)
        self.claims = create_access_token_claims(
            self.oauth_client,
            self.user,
            self.issuer,
            aud=[self.resource_server],
        )
        self.access_token = create_access_token(self.claims, self.jwks)

    def test_introspection(self):
        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/introspect", data={"token": self.access_token}, headers=headers
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertTrue(resp["active"])
        self.assertEqual(resp["client_id"], self.oauth_client.client_id)
        self.assertEqual(resp["token_type"], "Bearer")
        self.assertEqual(resp["scope"], self.oauth_client.scope)
        self.assertEqual(resp["sub"], self.user.id)
        self.assertEqual(resp["aud"], [self.resource_server])
        self.assertEqual(resp["iss"], self.issuer)

    def test_introspection_username(self):
        self.introspection_endpoint.get_username = lambda user_id: db.session.get(
            User, user_id
        ).username

        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/introspect", data={"token": self.access_token}, headers=headers
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertTrue(resp["active"])
        self.assertEqual(resp["username"], self.user.username)

    def test_non_access_token_skipped(self):
        class MyIntrospectionEndpoint(IntrospectionEndpoint):
            def query_token(self, token, token_type_hint):
                return None

        self.authorization_server.register_endpoint(MyIntrospectionEndpoint)
        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/introspect",
            data={
                "token": "refresh-token",
                "token_type_hint": "refresh_token",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertFalse(resp["active"])

    def test_access_token_non_jwt_skipped(self):
        class MyIntrospectionEndpoint(IntrospectionEndpoint):
            def query_token(self, token, token_type_hint):
                return None

        self.authorization_server.register_endpoint(MyIntrospectionEndpoint)
        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/introspect",
            data={
                "token": "non-jwt-access-token",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertFalse(resp["active"])

    def test_permission_denied(self):
        self.introspection_endpoint.check_permission = lambda *args: False

        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/introspect", data={"token": self.access_token}, headers=headers
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertFalse(resp["active"])

    def test_token_expired(self):
        self.claims["exp"] = time.time() - 3600
        access_token = create_access_token(self.claims, self.jwks)
        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/introspect", data={"token": access_token}, headers=headers
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertFalse(resp["active"])

    def test_introspection_different_issuer(self):
        class MyIntrospectionEndpoint(IntrospectionEndpoint):
            def query_token(self, token, token_type_hint):
                return None

        self.authorization_server.register_endpoint(MyIntrospectionEndpoint)

        self.claims["iss"] = "different-issuer"
        access_token = create_access_token(self.claims, self.jwks)
        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/introspect", data={"token": access_token}, headers=headers
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertFalse(resp["active"])

    def test_introspection_invalid_claim(self):
        self.claims["exp"] = "invalid"
        access_token = create_access_token(self.claims, self.jwks)
        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/introspect", data={"token": access_token}, headers=headers
        )
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")


class JWTAccessTokenRevocationTest(TestCase):
    def setUp(self):
        super().setUp()
        self.issuer = "https://authlib.org/"
        self.resource_server = "resource-server-id"
        self.jwks = read_file_path("jwks_private.json")
        self.authorization_server = create_authorization_server(self.app)
        self.authorization_server.register_grant(AuthorizationCodeGrant)
        self.revocation_endpoint = create_revocation_endpoint(
            self.app, self.authorization_server, self.issuer, self.jwks
        )
        self.user = create_user()
        self.oauth_client = create_oauth_client("client-id", self.user)
        self.claims = create_access_token_claims(
            self.oauth_client,
            self.user,
            self.issuer,
            aud=[self.resource_server],
        )
        self.access_token = create_access_token(self.claims, self.jwks)

    def test_revocation(self):
        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/revoke", data={"token": self.access_token}, headers=headers
        )
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unsupported_token_type")

    def test_non_access_token_skipped(self):
        class MyRevocationEndpoint(RevocationEndpoint):
            def query_token(self, token, token_type_hint):
                return None

        self.authorization_server.register_endpoint(MyRevocationEndpoint)
        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/revoke",
            data={
                "token": "refresh-token",
                "token_type_hint": "refresh_token",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertEqual(resp, {})

    def test_access_token_non_jwt_skipped(self):
        class MyRevocationEndpoint(RevocationEndpoint):
            def query_token(self, token, token_type_hint):
                return None

        self.authorization_server.register_endpoint(MyRevocationEndpoint)
        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/revoke",
            data={
                "token": "non-jwt-access-token",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertEqual(resp, {})

    def test_revocation_different_issuer(self):
        self.claims["iss"] = "different-issuer"
        access_token = create_access_token(self.claims, self.jwks)

        headers = self.create_basic_header(
            self.oauth_client.client_id, self.oauth_client.client_secret
        )
        rv = self.client.post(
            "/oauth/revoke", data={"token": access_token}, headers=headers
        )
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unsupported_token_type")



from django.test import override_settings

from authlib.oauth1.rfc5849 import errors
from tests.util import decode_response

from .models import Client
from .models import User
from .oauth1_server import TestCase


class AuthorizationTest(TestCase):
    def prepare_data(self):
        user = User(username="foo")
        user.save()
        client = Client(
            user_id=user.pk,
            client_id="client",
            client_secret="secret",
            default_redirect_uri="https://a.b",
        )
        client.save()

    def test_invalid_authorization(self):
        server = self.create_server()
        url = "/oauth/authorize"
        request = self.factory.post(url)
        self.assertRaises(
            errors.MissingRequiredParameterError,
            server.check_authorization_request,
            request,
        )

        request = self.factory.post(url, data={"oauth_token": "a"})
        self.assertRaises(
            errors.InvalidTokenError, server.check_authorization_request, request
        )

    def test_invalid_initiate(self):
        server = self.create_server()
        url = "/oauth/initiate"
        request = self.factory.post(
            url,
            data={
                "oauth_consumer_key": "client",
                "oauth_callback": "oob",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_signature": "secret&",
            },
        )
        resp = server.create_temporary_credentials_response(request)
        data = decode_response(resp.content)
        self.assertEqual(data["error"], "invalid_client")

    @override_settings(AUTHLIB_OAUTH1_PROVIDER={"signature_methods": ["PLAINTEXT"]})
    def test_authorize_denied(self):
        self.prepare_data()
        server = self.create_server()
        initiate_url = "/oauth/initiate"
        authorize_url = "/oauth/authorize"

        # case 1
        request = self.factory.post(
            initiate_url,
            data={
                "oauth_consumer_key": "client",
                "oauth_callback": "oob",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_signature": "secret&",
            },
        )
        resp = server.create_temporary_credentials_response(request)
        data = decode_response(resp.content)
        self.assertIn("oauth_token", data)

        request = self.factory.post(
            authorize_url, data={"oauth_token": data["oauth_token"]}
        )
        resp = server.create_authorization_response(request)
        self.assertEqual(resp.status_code, 302)
        self.assertIn("access_denied", resp["Location"])
        self.assertIn("https://a.b", resp["Location"])

        # case 2
        request = self.factory.post(
            initiate_url,
            data={
                "oauth_consumer_key": "client",
                "oauth_callback": "https://i.test",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_signature": "secret&",
            },
        )
        resp = server.create_temporary_credentials_response(request)
        data = decode_response(resp.content)
        self.assertIn("oauth_token", data)
        request = self.factory.post(
            authorize_url, data={"oauth_token": data["oauth_token"]}
        )
        resp = server.create_authorization_response(request)
        self.assertEqual(resp.status_code, 302)
        self.assertIn("access_denied", resp["Location"])
        self.assertIn("https://i.test", resp["Location"])

    @override_settings(AUTHLIB_OAUTH1_PROVIDER={"signature_methods": ["PLAINTEXT"]})
    def test_authorize_granted(self):
        self.prepare_data()
        server = self.create_server()
        user = User.objects.get(username="foo")
        initiate_url = "/oauth/initiate"
        authorize_url = "/oauth/authorize"

        # case 1
        request = self.factory.post(
            initiate_url,
            data={
                "oauth_consumer_key": "client",
                "oauth_callback": "oob",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_signature": "secret&",
            },
        )
        resp = server.create_temporary_credentials_response(request)
        data = decode_response(resp.content)
        self.assertIn("oauth_token", data)

        request = self.factory.post(
            authorize_url, data={"oauth_token": data["oauth_token"]}
        )
        resp = server.create_authorization_response(request, user)
        self.assertEqual(resp.status_code, 302)

        self.assertIn("oauth_verifier", resp["Location"])
        self.assertIn("https://a.b", resp["Location"])

        # case 2
        request = self.factory.post(
            initiate_url,
            data={
                "oauth_consumer_key": "client",
                "oauth_callback": "https://i.test",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_signature": "secret&",
            },
        )
        resp = server.create_temporary_credentials_response(request)
        data = decode_response(resp.content)
        self.assertIn("oauth_token", data)

        request = self.factory.post(
            authorize_url, data={"oauth_token": data["oauth_token"]}
        )
        resp = server.create_authorization_response(request, user)

        self.assertEqual(resp.status_code, 302)
        self.assertIn("oauth_verifier", resp["Location"])
        self.assertIn("https://i.test", resp["Location"])



from flask import json

from authlib.oauth2.rfc7592 import (
    ClientConfigurationEndpoint as _ClientConfigurationEndpoint,
)

from .models import Client
from .models import Token
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class ClientConfigurationEndpoint(_ClientConfigurationEndpoint):
    software_statement_alg_values_supported = ["RS256"]

    def authenticate_token(self, request):
        auth_header = request.headers.get("Authorization")
        if auth_header:
            access_token = auth_header.split()[1]
            return Token.query.filter_by(access_token=access_token).first()

    def update_client(self, client, client_metadata, request):
        client.set_client_metadata(client_metadata)
        db.session.add(client)
        db.session.commit()
        return client

    def authenticate_client(self, request):
        client_id = request.uri.split("/")[-1]
        return Client.query.filter_by(client_id=client_id).first()

    def revoke_access_token(self, request, token):
        token.revoked = True
        db.session.add(token)
        db.session.commit()

    def check_permission(self, client, request):
        client_id = request.uri.split("/")[-1]
        return client_id != "unauthorized_client_id"

    def delete_client(self, client, request):
        db.session.delete(client)
        db.session.commit()

    def generate_client_registration_info(self, client, request):
        return {
            "registration_client_uri": request.uri,
            "registration_access_token": request.headers["Authorization"].split(" ")[1],
        }


class ClientConfigurationTestMixin(TestCase):
    def prepare_data(self, endpoint_cls=None, metadata=None):
        app = self.app
        server = create_authorization_server(app)

        if endpoint_cls:
            server.register_endpoint(endpoint_cls)
        else:

            class MyClientConfiguration(ClientConfigurationEndpoint):
                def get_server_metadata(self):
                    return metadata

            server.register_endpoint(MyClientConfiguration)

        @app.route("/configure_client/<client_id>", methods=["PUT", "GET", "DELETE"])
        def configure_client(client_id):
            return server.create_endpoint_response(
                ClientConfigurationEndpoint.ENDPOINT_NAME
            )

        user = User(username="foo")
        db.session.add(user)

        client = Client(
            client_id="client_id",
            client_secret="client_secret",
        )
        client.set_client_metadata(
            {
                "client_name": "Authlib",
                "scope": "openid profile",
            }
        )
        db.session.add(client)

        token = Token(
            user_id=user.id,
            client_id=client.id,
            token_type="bearer",
            access_token="a1",
            refresh_token="r1",
            scope="openid profile",
            expires_in=3600,
        )
        db.session.add(token)

        db.session.commit()
        return user, client, token


class ClientConfigurationReadTest(ClientConfigurationTestMixin):
    def test_read_client(self):
        user, client, token = self.prepare_data()
        self.assertEqual(client.client_name, "Authlib")
        headers = {"Authorization": f"bearer {token.access_token}"}
        rv = self.client.get("/configure_client/client_id", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(resp["client_id"], client.client_id)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(
            resp["registration_client_uri"],
            "http://localhost/configure_client/client_id",
        )
        self.assertEqual(resp["registration_access_token"], token.access_token)

    def test_access_denied(self):
        user, client, token = self.prepare_data()
        rv = self.client.get("/configure_client/client_id")
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "access_denied")

        headers = {"Authorization": "bearer invalid_token"}
        rv = self.client.get("/configure_client/client_id", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "access_denied")

        headers = {"Authorization": "bearer unauthorized_token"}
        rv = self.client.get(
            "/configure_client/client_id",
            json={"client_id": "client_id", "client_name": "new client_name"},
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "access_denied")

    def test_invalid_client(self):
        # If the client does not exist on this server, the server MUST respond
        # with HTTP 401 Unauthorized, and the registration access token used to
        # make this request SHOULD be immediately revoked.
        user, client, token = self.prepare_data()

        headers = {"Authorization": f"bearer {token.access_token}"}
        rv = self.client.get("/configure_client/invalid_client_id", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 401)
        self.assertEqual(resp["error"], "invalid_client")

    def test_unauthorized_client(self):
        # If the client does not have permission to read its record, the server
        # MUST return an HTTP 403 Forbidden.

        client = Client(
            client_id="unauthorized_client_id",
            client_secret="unauthorized_client_secret",
        )
        db.session.add(client)

        user, client, token = self.prepare_data()

        headers = {"Authorization": f"bearer {token.access_token}"}
        rv = self.client.get(
            "/configure_client/unauthorized_client_id", headers=headers
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 403)
        self.assertEqual(resp["error"], "unauthorized_client")


class ClientConfigurationUpdateTest(ClientConfigurationTestMixin):
    def test_update_client(self):
        # Valid values of client metadata fields in this request MUST replace,
        # not augment, the values previously associated with this client.
        # Omitted fields MUST be treated as null or empty values by the server,
        # indicating the client's request to delete them from the client's
        # registration.  The authorization server MAY ignore any null or empty
        # value in the request just as any other value.

        user, client, token = self.prepare_data()
        self.assertEqual(client.client_name, "Authlib")
        headers = {"Authorization": f"bearer {token.access_token}"}
        body = {
            "client_id": client.client_id,
            "client_name": "NewAuthlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(resp["client_id"], client.client_id)
        self.assertEqual(resp["client_name"], "NewAuthlib")
        self.assertEqual(client.client_name, "NewAuthlib")
        self.assertEqual(client.scope, "")

    def test_access_denied(self):
        user, client, token = self.prepare_data()
        rv = self.client.put("/configure_client/client_id", json={})
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "access_denied")

        headers = {"Authorization": "bearer invalid_token"}
        rv = self.client.put("/configure_client/client_id", json={}, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "access_denied")

        headers = {"Authorization": "bearer unauthorized_token"}
        rv = self.client.put(
            "/configure_client/client_id",
            json={"client_id": "client_id", "client_name": "new client_name"},
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "access_denied")

    def test_invalid_request(self):
        user, client, token = self.prepare_data()
        headers = {"Authorization": f"bearer {token.access_token}"}

        # The client MUST include its 'client_id' field in the request...
        rv = self.client.put("/configure_client/client_id", json={}, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "invalid_request")

        # ... and it MUST be the same as its currently issued client identifier.
        rv = self.client.put(
            "/configure_client/client_id",
            json={"client_id": "invalid_client_id"},
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "invalid_request")

        # The updated client metadata fields request MUST NOT include the
        # 'registration_access_token', 'registration_client_uri',
        # 'client_secret_expires_at', or 'client_id_issued_at' fields
        rv = self.client.put(
            "/configure_client/client_id",
            json={
                "client_id": "client_id",
                "registration_client_uri": "https://foobar.com",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "invalid_request")

        # If the client includes the 'client_secret' field in the request,
        # the value of this field MUST match the currently issued client
        # secret for that client.
        rv = self.client.put(
            "/configure_client/client_id",
            json={"client_id": "client_id", "client_secret": "invalid_secret"},
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "invalid_request")

    def test_invalid_client(self):
        # If the client does not exist on this server, the server MUST respond
        # with HTTP 401 Unauthorized, and the registration access token used to
        # make this request SHOULD be immediately revoked.
        user, client, token = self.prepare_data()

        headers = {"Authorization": f"bearer {token.access_token}"}
        rv = self.client.put(
            "/configure_client/invalid_client_id",
            json={"client_id": "invalid_client_id", "client_name": "new client_name"},
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 401)
        self.assertEqual(resp["error"], "invalid_client")

    def test_unauthorized_client(self):
        # If the client does not have permission to read its record, the server
        # MUST return an HTTP 403 Forbidden.

        client = Client(
            client_id="unauthorized_client_id",
            client_secret="unauthorized_client_secret",
        )
        db.session.add(client)

        user, client, token = self.prepare_data()

        headers = {"Authorization": f"bearer {token.access_token}"}
        rv = self.client.put(
            "/configure_client/unauthorized_client_id",
            json={
                "client_id": "unauthorized_client_id",
                "client_name": "new client_name",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 403)
        self.assertEqual(resp["error"], "unauthorized_client")

    def test_invalid_metadata(self):
        metadata = {"token_endpoint_auth_methods_supported": ["client_secret_basic"]}
        user, client, token = self.prepare_data(metadata=metadata)
        headers = {"Authorization": f"bearer {token.access_token}"}

        # For all metadata fields, the authorization server MAY replace any
        # invalid values with suitable default values, and it MUST return any
        # such fields to the client in the response.
        # If the client attempts to set an invalid metadata field and the
        # authorization server does not set a default value, the authorization
        # server responds with an error as described in [RFC7591].

        body = {
            "client_id": client.client_id,
            "client_name": "NewAuthlib",
            "token_endpoint_auth_method": "invalid_auth_method",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "invalid_client_metadata")

    def test_scopes_supported(self):
        metadata = {"scopes_supported": ["profile", "email"]}
        user, client, token = self.prepare_data(metadata=metadata)

        headers = {"Authorization": f"bearer {token.access_token}"}
        body = {
            "client_id": "client_id",
            "scope": "profile email",
            "client_name": "Authlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["client_id"], "client_id")
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["scope"], "profile email")

        headers = {"Authorization": f"bearer {token.access_token}"}
        body = {
            "client_id": "client_id",
            "scope": "",
            "client_name": "Authlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["client_id"], "client_id")
        self.assertEqual(resp["client_name"], "Authlib")

        body = {
            "client_id": "client_id",
            "scope": "profile email address",
            "client_name": "Authlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_response_types_supported(self):
        metadata = {"response_types_supported": ["code"]}
        user, client, token = self.prepare_data(metadata=metadata)

        headers = {"Authorization": f"bearer {token.access_token}"}
        body = {
            "client_id": "client_id",
            "response_types": ["code"],
            "client_name": "Authlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["client_id"], "client_id")
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["response_types"], ["code"])

        # https://datatracker.ietf.org/doc/html/rfc7592#section-2.2
        # If omitted, the default is that the client will use only the "code"
        # response type.
        body = {"client_id": "client_id", "client_name": "Authlib"}
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertNotIn("response_types", resp)

        body = {
            "client_id": "client_id",
            "response_types": ["code", "token"],
            "client_name": "Authlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_grant_types_supported(self):
        metadata = {"grant_types_supported": ["authorization_code", "password"]}
        user, client, token = self.prepare_data(metadata=metadata)

        headers = {"Authorization": f"bearer {token.access_token}"}
        body = {
            "client_id": "client_id",
            "grant_types": ["password"],
            "client_name": "Authlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["client_id"], "client_id")
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["grant_types"], ["password"])

        # https://datatracker.ietf.org/doc/html/rfc7592#section-2.2
        # If omitted, the default behavior is that the client will use only
        # the "authorization_code" Grant Type.
        body = {"client_id": "client_id", "client_name": "Authlib"}
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertNotIn("grant_types", resp)

        body = {
            "client_id": "client_id",
            "grant_types": ["client_credentials"],
            "client_name": "Authlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_token_endpoint_auth_methods_supported(self):
        metadata = {"token_endpoint_auth_methods_supported": ["client_secret_basic"]}
        user, client, token = self.prepare_data(metadata=metadata)

        headers = {"Authorization": f"bearer {token.access_token}"}
        body = {
            "client_id": "client_id",
            "token_endpoint_auth_method": "client_secret_basic",
            "client_name": "Authlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["client_id"], "client_id")
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["token_endpoint_auth_method"], "client_secret_basic")

        body = {
            "client_id": "client_id",
            "token_endpoint_auth_method": "none",
            "client_name": "Authlib",
        }
        rv = self.client.put("/configure_client/client_id", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")


class ClientConfigurationDeleteTest(ClientConfigurationTestMixin):
    def test_delete_client(self):
        user, client, token = self.prepare_data()
        self.assertEqual(client.client_name, "Authlib")
        headers = {"Authorization": f"bearer {token.access_token}"}
        rv = self.client.delete("/configure_client/client_id", headers=headers)
        self.assertEqual(rv.status_code, 204)
        self.assertFalse(rv.data)

    def test_access_denied(self):
        user, client, token = self.prepare_data()
        rv = self.client.delete("/configure_client/client_id")
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "access_denied")

        headers = {"Authorization": "bearer invalid_token"}
        rv = self.client.delete("/configure_client/client_id", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "access_denied")

        headers = {"Authorization": "bearer unauthorized_token"}
        rv = self.client.delete(
            "/configure_client/client_id",
            json={"client_id": "client_id", "client_name": "new client_name"},
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(resp["error"], "access_denied")

    def test_invalid_client(self):
        # If the client does not exist on this server, the server MUST respond
        # with HTTP 401 Unauthorized, and the registration access token used to
        # make this request SHOULD be immediately revoked.
        user, client, token = self.prepare_data()

        headers = {"Authorization": f"bearer {token.access_token}"}
        rv = self.client.delete("/configure_client/invalid_client_id", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 401)
        self.assertEqual(resp["error"], "invalid_client")

    def test_unauthorized_client(self):
        # If the client does not have permission to read its record, the server
        # MUST return an HTTP 403 Forbidden.

        client = Client(
            client_id="unauthorized_client_id",
            client_secret="unauthorized_client_secret",
        )
        db.session.add(client)

        user, client, token = self.prepare_data()

        headers = {"Authorization": f"bearer {token.access_token}"}
        rv = self.client.delete(
            "/configure_client/unauthorized_client_id", headers=headers
        )
        resp = json.loads(rv.data)
        self.assertEqual(rv.status_code, 403)
        self.assertEqual(resp["error"], "unauthorized_client")



from collections import defaultdict

from authlib.common.encoding import json_loads
from authlib.common.urls import url_decode
from authlib.common.urls import urlparse

from .errors import InsecureTransportError


class OAuth2Request:
    def __init__(self, method: str, uri: str, body=None, headers=None):
        InsecureTransportError.check(uri)
        #: HTTP method
        self.method = method
        self.uri = uri
        self.body = body
        #: HTTP headers
        self.headers = headers or {}

        self.client = None
        self.auth_method = None
        self.user = None
        self.authorization_code = None
        self.refresh_token = None
        self.credential = None

        self._parsed_query = None

    @property
    def args(self):
        if self._parsed_query is None:
            self._parsed_query = url_decode(urlparse.urlparse(self.uri).query)
        return dict(self._parsed_query)

    @property
    def form(self):
        return self.body or {}

    @property
    def data(self):
        data = {}
        data.update(self.args)
        data.update(self.form)
        return data

    @property
    def datalist(self) -> defaultdict[str, list]:
        """Return all the data in query parameters and the body of the request as a dictionary
        with all the values in lists.
        """
        if self._parsed_query is None:
            self._parsed_query = url_decode(urlparse.urlparse(self.uri).query)
        values = defaultdict(list)
        for k, v in self._parsed_query:
            values[k].append(v)
        for k, v in self.form.items():
            values[k].append(v)
        return values

    @property
    def client_id(self) -> str:
        """The authorization server issues the registered client a client
        identifier -- a unique string representing the registration
        information provided by the client. The value is extracted from
        request.

        :return: string
        """
        return self.data.get("client_id")

    @property
    def response_type(self) -> str:
        rt = self.data.get("response_type")
        if rt and " " in rt:
            # sort multiple response types
            return " ".join(sorted(rt.split()))
        return rt

    @property
    def grant_type(self) -> str:
        return self.form.get("grant_type")

    @property
    def redirect_uri(self):
        return self.data.get("redirect_uri")

    @property
    def scope(self) -> str:
        return self.data.get("scope")

    @property
    def state(self):
        return self.data.get("state")


class JsonRequest:
    def __init__(self, method, uri, body=None, headers=None):
        self.method = method
        self.uri = uri
        self.body = body
        self.headers = headers or {}

    @property
    def data(self):
        return json_loads(self.body)



from flask import json
from flask import jsonify

from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.integrations.flask_oauth2 import current_token
from authlib.integrations.sqla_oauth2 import create_bearer_token_validator

from .models import Client
from .models import Token
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server

require_oauth = ResourceProtector()
BearerTokenValidator = create_bearer_token_validator(db.session, Token)
require_oauth.register_token_validator(BearerTokenValidator())


def create_resource_server(app):
    @app.route("/user")
    @require_oauth("profile")
    def user_profile():
        user = current_token.user
        return jsonify(id=user.id, username=user.username)

    @app.route("/user/email")
    @require_oauth("email")
    def user_email():
        user = current_token.user
        return jsonify(email=user.username + "@example.com")

    @app.route("/info")
    @require_oauth()
    def public_info():
        return jsonify(status="ok")

    @app.route("/operator-and")
    @require_oauth(["profile email"])
    def operator_and():
        return jsonify(status="ok")

    @app.route("/operator-or")
    @require_oauth(["profile", "email"])
    def operator_or():
        return jsonify(status="ok")

    @app.route("/acquire")
    def test_acquire():
        with require_oauth.acquire("profile") as token:
            user = token.user
            return jsonify(id=user.id, username=user.username)

    @app.route("/optional")
    @require_oauth("profile", optional=True)
    def test_optional_token():
        if current_token:
            user = current_token.user
            return jsonify(id=user.id, username=user.username)
        else:
            return jsonify(id=0, username="anonymous")


class AuthorizationTest(TestCase):
    def test_none_grant(self):
        create_authorization_server(self.app)
        authorize_url = "/oauth/authorize?response_type=token&client_id=implicit-client"
        rv = self.client.get(authorize_url)
        self.assertIn(b"unsupported_response_type", rv.data)

        rv = self.client.post(authorize_url, data={"user_id": "1"})
        self.assertNotEqual(rv.status, 200)

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": "x",
            },
        )
        data = json.loads(rv.data)
        self.assertEqual(data["error"], "unsupported_grant_type")


class ResourceTest(TestCase):
    def prepare_data(self):
        create_resource_server(self.app)

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="resource-client",
            client_secret="resource-secret",
        )
        client.set_client_metadata(
            {
                "scope": "profile",
                "redirect_uris": ["http://localhost/authorized"],
            }
        )
        db.session.add(client)
        db.session.commit()

    def create_token(self, expires_in=3600):
        token = Token(
            user_id=1,
            client_id="resource-client",
            token_type="bearer",
            access_token="a1",
            scope="profile",
            expires_in=expires_in,
        )
        db.session.add(token)
        db.session.commit()

    def create_bearer_header(self, token):
        return {"Authorization": "Bearer " + token}

    def test_invalid_token(self):
        self.prepare_data()

        rv = self.client.get("/user")
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "missing_authorization")

        headers = {"Authorization": "invalid token"}
        rv = self.client.get("/user", headers=headers)
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unsupported_token_type")

        headers = self.create_bearer_header("invalid")
        rv = self.client.get("/user", headers=headers)
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")

    def test_expired_token(self):
        self.prepare_data()
        self.create_token(-10)
        headers = self.create_bearer_header("a1")

        rv = self.client.get("/user", headers=headers)
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_token")

        rv = self.client.get("/acquire", headers=headers)
        self.assertEqual(rv.status_code, 401)

    def test_insufficient_token(self):
        self.prepare_data()
        self.create_token()
        headers = self.create_bearer_header("a1")
        rv = self.client.get("/user/email", headers=headers)
        self.assertEqual(rv.status_code, 403)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "insufficient_scope")

    def test_access_resource(self):
        self.prepare_data()
        self.create_token()
        headers = self.create_bearer_header("a1")

        rv = self.client.get("/user", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["username"], "foo")

        rv = self.client.get("/acquire", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["username"], "foo")

        rv = self.client.get("/info", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["status"], "ok")

    def test_scope_operator(self):
        self.prepare_data()
        self.create_token()
        headers = self.create_bearer_header("a1")
        rv = self.client.get("/operator-and", headers=headers)
        self.assertEqual(rv.status_code, 403)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "insufficient_scope")

        rv = self.client.get("/operator-or", headers=headers)
        self.assertEqual(rv.status_code, 200)

    def test_optional_token(self):
        self.prepare_data()
        rv = self.client.get("/optional")
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertEqual(resp["username"], "anonymous")

        self.create_token()
        headers = self.create_bearer_header("a1")
        rv = self.client.get("/optional", headers=headers)
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertEqual(resp["username"], "foo")



from unittest import TestCase

from authlib.jose.errors import InvalidClaimError
from authlib.oidc.registration import ClientMetadataClaims


class ClientMetadataClaimsTest(TestCase):
    def test_request_uris(self):
        claims = ClientMetadataClaims(
            {"request_uris": ["https://client.test/request_uris"]}, {}
        )
        claims.validate()

        claims = ClientMetadataClaims({"request_uris": ["invalid"]}, {})
        self.assertRaises(InvalidClaimError, claims.validate)

    def test_initiate_login_uri(self):
        claims = ClientMetadataClaims(
            {"initiate_login_uri": "https://client.test/initiate_login_uri"}, {}
        )
        claims.validate()

        claims = ClientMetadataClaims({"initiate_login_uri": "invalid"}, {})
        self.assertRaises(InvalidClaimError, claims.validate)

    def test_token_endpoint_auth_signing_alg(self):
        claims = ClientMetadataClaims({"token_endpoint_auth_signing_alg": "RSA256"}, {})
        claims.validate()

        # The value none MUST NOT be used.
        claims = ClientMetadataClaims({"token_endpoint_auth_signing_alg": "none"}, {})
        self.assertRaises(InvalidClaimError, claims.validate)

    def test_id_token_signed_response_alg(self):
        claims = ClientMetadataClaims({"id_token_signed_response_alg": "RSA256"}, {})
        claims.validate()

        # The value none MUST NOT be used.
        claims = ClientMetadataClaims({"id_token_signed_response_alg": "none"}, {})
        self.assertRaises(InvalidClaimError, claims.validate)

    def test_default_max_age(self):
        claims = ClientMetadataClaims({"default_max_age": 1234}, {})
        claims.validate()

        # The value none MUST NOT be used.
        claims = ClientMetadataClaims({"default_max_age": "invalid"}, {})
        self.assertRaises(InvalidClaimError, claims.validate)



SECRET_KEY = "django-secret"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

MIDDLEWARE = ["django.contrib.sessions.middleware.SessionMiddleware"]

SESSION_ENGINE = "django.contrib.sessions.backends.cache"

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "unique-snowflake",
    }
}

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "tests.django.test_oauth1",
    "tests.django.test_oauth2",
]

USE_TZ = True



from unittest import TestCase

from authlib.jose.errors import InvalidClaimError
from authlib.oauth2.rfc7591 import ClientMetadataClaims


class ClientMetadataClaimsTest(TestCase):
    def test_validate_redirect_uris(self):
        claims = ClientMetadataClaims({"redirect_uris": ["foo"]}, {})
        self.assertRaises(InvalidClaimError, claims.validate)

    def test_validate_client_uri(self):
        claims = ClientMetadataClaims({"client_uri": "foo"}, {})
        self.assertRaises(InvalidClaimError, claims.validate)

    def test_validate_logo_uri(self):
        claims = ClientMetadataClaims({"logo_uri": "foo"}, {})
        self.assertRaises(InvalidClaimError, claims.validate)

    def test_validate_tos_uri(self):
        claims = ClientMetadataClaims({"tos_uri": "foo"}, {})
        self.assertRaises(InvalidClaimError, claims.validate)

    def test_validate_policy_uri(self):
        claims = ClientMetadataClaims({"policy_uri": "foo"}, {})
        self.assertRaises(InvalidClaimError, claims.validate)

    def test_validate_jwks_uri(self):
        claims = ClientMetadataClaims({"jwks_uri": "foo"}, {})
        self.assertRaises(InvalidClaimError, claims.validate)



from unittest import TestCase
from unittest import mock

from flask import Flask

from authlib.integrations.flask_client import OAuth
from authlib.jose import JsonWebKey
from authlib.jose.errors import InvalidClaimError
from authlib.oidc.core.grants.util import generate_id_token

from ..util import get_bearer_token
from ..util import read_key_file

secret_key = JsonWebKey.import_key("secret", {"kty": "oct", "kid": "f"})


class FlaskUserMixinTest(TestCase):
    def test_fetch_userinfo(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            fetch_token=get_bearer_token,
            userinfo_endpoint="https://i.b/userinfo",
        )

        def fake_send(sess, req, **kwargs):
            resp = mock.MagicMock()
            resp.json = lambda: {"sub": "123"}
            resp.status_code = 200
            return resp

        with app.test_request_context():
            with mock.patch("requests.sessions.Session.send", fake_send):
                user = client.userinfo()
                self.assertEqual(user.sub, "123")

    def test_parse_id_token(self):
        token = get_bearer_token()
        id_token = generate_id_token(
            token,
            {"sub": "123"},
            secret_key,
            alg="HS256",
            iss="https://i.b",
            aud="dev",
            exp=3600,
            nonce="n",
        )

        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            fetch_token=get_bearer_token,
            jwks={"keys": [secret_key.as_dict()]},
            issuer="https://i.b",
            id_token_signing_alg_values_supported=["HS256", "RS256"],
        )
        with app.test_request_context():
            self.assertIsNone(client.parse_id_token(token, nonce="n"))

            token["id_token"] = id_token
            user = client.parse_id_token(token, nonce="n")
            self.assertEqual(user.sub, "123")

            claims_options = {"iss": {"value": "https://i.b"}}
            user = client.parse_id_token(
                token, nonce="n", claims_options=claims_options
            )
            self.assertEqual(user.sub, "123")

            claims_options = {"iss": {"value": "https://i.c"}}
            self.assertRaises(
                InvalidClaimError, client.parse_id_token, token, "n", claims_options
            )

    def test_parse_id_token_nonce_supported(self):
        token = get_bearer_token()
        id_token = generate_id_token(
            token,
            {"sub": "123", "nonce_supported": False},
            secret_key,
            alg="HS256",
            iss="https://i.b",
            aud="dev",
            exp=3600,
        )

        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            fetch_token=get_bearer_token,
            jwks={"keys": [secret_key.as_dict()]},
            issuer="https://i.b",
            id_token_signing_alg_values_supported=["HS256", "RS256"],
        )
        with app.test_request_context():
            token["id_token"] = id_token
            user = client.parse_id_token(token, nonce="n")
            self.assertEqual(user.sub, "123")

    def test_runtime_error_fetch_jwks_uri(self):
        token = get_bearer_token()
        id_token = generate_id_token(
            token,
            {"sub": "123"},
            secret_key,
            alg="HS256",
            iss="https://i.b",
            aud="dev",
            exp=3600,
            nonce="n",
        )

        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        alt_key = secret_key.as_dict()
        alt_key["kid"] = "b"
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            fetch_token=get_bearer_token,
            jwks={"keys": [alt_key]},
            issuer="https://i.b",
            id_token_signing_alg_values_supported=["HS256"],
        )
        with app.test_request_context():
            token["id_token"] = id_token
            self.assertRaises(RuntimeError, client.parse_id_token, token, "n")

    def test_force_fetch_jwks_uri(self):
        secret_keys = read_key_file("jwks_private.json")
        token = get_bearer_token()
        id_token = generate_id_token(
            token,
            {"sub": "123"},
            secret_keys,
            alg="RS256",
            iss="https://i.b",
            aud="dev",
            exp=3600,
            nonce="n",
        )

        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            fetch_token=get_bearer_token,
            jwks={"keys": [secret_key.as_dict()]},
            jwks_uri="https://i.b/jwks",
            issuer="https://i.b",
        )

        def fake_send(sess, req, **kwargs):
            resp = mock.MagicMock()
            resp.json = lambda: read_key_file("jwks_public.json")
            resp.status_code = 200
            return resp

        with app.test_request_context():
            self.assertIsNone(client.parse_id_token(token, nonce="n"))

            with mock.patch("requests.sessions.Session.send", fake_send):
                token["id_token"] = id_token
                user = client.parse_id_token(token, nonce="n")
                self.assertEqual(user.sub, "123")



import unittest

from authlib.jose import JsonWebSignature
from tests.util import read_file_path


class EdDSATest(unittest.TestCase):
    def test_EdDSA_alg(self):
        jws = JsonWebSignature(algorithms=["EdDSA"])
        private_key = read_file_path("ed25519-pkcs8.pem")
        public_key = read_file_path("ed25519-pub.pem")
        s = jws.serialize({"alg": "EdDSA"}, "hello", private_key)
        data = jws.deserialize(s, public_key)
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header["alg"], "EdDSA")



from flask import json

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.jose import jwt
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oidc.core import HybridIDToken
from authlib.oidc.core.grants import OpenIDCode as _OpenIDCode
from authlib.oidc.core.grants import OpenIDHybridGrant as _OpenIDHybridGrant

from .models import Client
from .models import CodeGrantMixin
from .models import User
from .models import db
from .models import exists_nonce
from .models import save_authorization_code
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server

JWT_CONFIG = {"iss": "Authlib", "key": "secret", "alg": "HS256", "exp": 3600}


class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


class OpenIDCode(_OpenIDCode):
    def get_jwt_config(self, grant):
        return dict(JWT_CONFIG)

    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)

    def generate_user_info(self, user, scopes):
        return user.generate_user_info(scopes)


class OpenIDHybridGrant(_OpenIDHybridGrant):
    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)

    def get_jwt_config(self):
        return dict(JWT_CONFIG)

    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)

    def generate_user_info(self, user, scopes):
        return user.generate_user_info(scopes)


class OpenIDCodeTest(TestCase):
    def prepare_data(self):
        server = create_authorization_server(self.app)
        server.register_grant(OpenIDHybridGrant)
        server.register_grant(AuthorizationCodeGrant, [OpenIDCode()])

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()

        client = Client(
            user_id=user.id,
            client_id="hybrid-client",
            client_secret="hybrid-secret",
        )
        client.set_client_metadata(
            {
                "redirect_uris": ["https://a.b"],
                "scope": "openid profile address",
                "response_types": [
                    "code id_token",
                    "code token",
                    "code id_token token",
                ],
                "grant_types": ["authorization_code"],
            }
        )
        db.session.add(client)
        db.session.commit()

    def validate_claims(self, id_token, params):
        claims = jwt.decode(
            id_token, "secret", claims_cls=HybridIDToken, claims_params=params
        )
        claims.validate()

    def test_invalid_client_id(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "code token",
                "state": "bar",
                "nonce": "abc",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "invalid-client",
                "response_type": "code token",
                "state": "bar",
                "nonce": "abc",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_require_nonce(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "hybrid-client",
                "response_type": "code token",
                "scope": "openid profile",
                "state": "bar",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn("error=invalid_request", rv.location)
        self.assertIn("nonce", rv.location)

    def test_invalid_response_type(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "hybrid-client",
                "response_type": "code id_token invalid",
                "state": "bar",
                "nonce": "abc",
                "scope": "profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unsupported_response_type")

    def test_invalid_scope(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "hybrid-client",
                "response_type": "code id_token",
                "state": "bar",
                "nonce": "abc",
                "scope": "profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn("error=invalid_scope", rv.location)

    def test_access_denied(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "hybrid-client",
                "response_type": "code token",
                "state": "bar",
                "nonce": "abc",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
            },
        )
        self.assertIn("error=access_denied", rv.location)

    def test_code_access_token(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "hybrid-client",
                "response_type": "code token",
                "state": "bar",
                "nonce": "abc",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn("code=", rv.location)
        self.assertIn("access_token=", rv.location)
        self.assertNotIn("id_token=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
        self.assertEqual(params["state"], "bar")

        code = params["code"]
        headers = self.create_basic_header("hybrid-client", "hybrid-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "redirect_uri": "https://a.b",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("id_token", resp)

    def test_code_id_token(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "hybrid-client",
                "response_type": "code id_token",
                "state": "bar",
                "nonce": "abc",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn("code=", rv.location)
        self.assertIn("id_token=", rv.location)
        self.assertNotIn("access_token=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
        self.assertEqual(params["state"], "bar")

        params["nonce"] = "abc"
        params["client_id"] = "hybrid-client"
        self.validate_claims(params["id_token"], params)

        code = params["code"]
        headers = self.create_basic_header("hybrid-client", "hybrid-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "redirect_uri": "https://a.b",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("id_token", resp)

    def test_code_id_token_access_token(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "hybrid-client",
                "response_type": "code id_token token",
                "state": "bar",
                "nonce": "abc",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn("code=", rv.location)
        self.assertIn("id_token=", rv.location)
        self.assertIn("access_token=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
        self.assertEqual(params["state"], "bar")
        self.validate_claims(params["id_token"], params)

        code = params["code"]
        headers = self.create_basic_header("hybrid-client", "hybrid-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "redirect_uri": "https://a.b",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("id_token", resp)

    def test_response_mode_query(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "hybrid-client",
                "response_type": "code id_token token",
                "response_mode": "query",
                "state": "bar",
                "nonce": "abc",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn("code=", rv.location)
        self.assertIn("id_token=", rv.location)
        self.assertIn("access_token=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        self.assertEqual(params["state"], "bar")

    def test_response_mode_form_post(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "client_id": "hybrid-client",
                "response_type": "code id_token token",
                "response_mode": "form_post",
                "state": "bar",
                "nonce": "abc",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn(b'name="code"', rv.data)
        self.assertIn(b'name="id_token"', rv.data)
        self.assertIn(b'name="access_token"', rv.data)



import base64
import os

from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_unicode
from authlib.integrations.django_oauth2 import AuthorizationServer
from tests.django_helper import TestCase as _TestCase

from .models import Client
from .models import OAuth2Token


class TestCase(_TestCase):
    def setUp(self):
        super().setUp()
        os.environ["AUTHLIB_INSECURE_TRANSPORT"] = "true"

    def tearDown(self):
        super().tearDown()
        os.environ.pop("AUTHLIB_INSECURE_TRANSPORT")

    def create_server(self):
        return AuthorizationServer(Client, OAuth2Token)

    def create_basic_auth(self, username, password):
        text = f"{username}:{password}"
        auth = to_unicode(base64.b64encode(to_bytes(text)))
        return "Basic " + auth



import json
import unittest

from authlib.jose import JsonWebSignature
from authlib.jose import errors
from tests.util import read_file_path


class JWSTest(unittest.TestCase):
    def test_invalid_input(self):
        jws = JsonWebSignature()
        self.assertRaises(errors.DecodeError, jws.deserialize, "a", "k")
        self.assertRaises(errors.DecodeError, jws.deserialize, "a.b.c", "k")
        self.assertRaises(errors.DecodeError, jws.deserialize, "YQ.YQ.YQ", "k")  # a
        self.assertRaises(errors.DecodeError, jws.deserialize, "W10.a.YQ", "k")  # []
        self.assertRaises(errors.DecodeError, jws.deserialize, "e30.a.YQ", "k")  # {}
        self.assertRaises(
            errors.DecodeError, jws.deserialize, "eyJhbGciOiJzIn0.a.YQ", "k"
        )
        self.assertRaises(
            errors.DecodeError, jws.deserialize, "eyJhbGciOiJzIn0.YQ.a", "k"
        )

    def test_invalid_alg(self):
        jws = JsonWebSignature()
        self.assertRaises(
            errors.UnsupportedAlgorithmError,
            jws.deserialize,
            "eyJhbGciOiJzIn0.YQ.YQ",
            "k",
        )
        self.assertRaises(errors.MissingAlgorithmError, jws.serialize, {}, "", "k")
        self.assertRaises(
            errors.UnsupportedAlgorithmError, jws.serialize, {"alg": "s"}, "", "k"
        )

    def test_bad_signature(self):
        jws = JsonWebSignature()
        s = "eyJhbGciOiJIUzI1NiJ9.YQ.YQ"
        self.assertRaises(errors.BadSignatureError, jws.deserialize, s, "k")

    def test_not_supported_alg(self):
        jws = JsonWebSignature(algorithms=["HS256"])
        s = jws.serialize({"alg": "HS256"}, "hello", "secret")

        jws = JsonWebSignature(algorithms=["RS256"])
        self.assertRaises(
            errors.UnsupportedAlgorithmError,
            lambda: jws.serialize({"alg": "HS256"}, "hello", "secret"),
        )

        self.assertRaises(
            errors.UnsupportedAlgorithmError, jws.deserialize, s, "secret"
        )

    def test_compact_jws(self):
        jws = JsonWebSignature(algorithms=["HS256"])
        s = jws.serialize({"alg": "HS256"}, "hello", "secret")
        data = jws.deserialize(s, "secret")
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header["alg"], "HS256")
        self.assertNotIn("signature", data)

    def test_compact_rsa(self):
        jws = JsonWebSignature()
        private_key = read_file_path("rsa_private.pem")
        public_key = read_file_path("rsa_public.pem")
        s = jws.serialize({"alg": "RS256"}, "hello", private_key)
        data = jws.deserialize(s, public_key)
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header["alg"], "RS256")

        # can deserialize with private key
        data2 = jws.deserialize(s, private_key)
        self.assertEqual(data, data2)

        ssh_pub_key = read_file_path("ssh_public.pem")
        self.assertRaises(errors.BadSignatureError, jws.deserialize, s, ssh_pub_key)

    def test_compact_rsa_pss(self):
        jws = JsonWebSignature()
        private_key = read_file_path("rsa_private.pem")
        public_key = read_file_path("rsa_public.pem")
        s = jws.serialize({"alg": "PS256"}, "hello", private_key)
        data = jws.deserialize(s, public_key)
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header["alg"], "PS256")
        ssh_pub_key = read_file_path("ssh_public.pem")
        self.assertRaises(errors.BadSignatureError, jws.deserialize, s, ssh_pub_key)

    def test_compact_none(self):
        jws = JsonWebSignature()
        s = jws.serialize({"alg": "none"}, "hello", "")
        self.assertRaises(errors.BadSignatureError, jws.deserialize, s, "")

    def test_flattened_json_jws(self):
        jws = JsonWebSignature()
        protected = {"alg": "HS256"}
        header = {"protected": protected, "header": {"kid": "a"}}
        s = jws.serialize(header, "hello", "secret")
        self.assertIsInstance(s, dict)

        data = jws.deserialize(s, "secret")
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header["alg"], "HS256")
        self.assertNotIn("protected", data)

    def test_nested_json_jws(self):
        jws = JsonWebSignature()
        protected = {"alg": "HS256"}
        header = {"protected": protected, "header": {"kid": "a"}}
        s = jws.serialize([header], "hello", "secret")
        self.assertIsInstance(s, dict)
        self.assertIn("signatures", s)

        data = jws.deserialize(s, "secret")
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header[0]["alg"], "HS256")
        self.assertNotIn("signatures", data)

        # test bad signature
        self.assertRaises(errors.BadSignatureError, jws.deserialize, s, "f")

    def test_function_key(self):
        protected = {"alg": "HS256"}
        header = [
            {"protected": protected, "header": {"kid": "a"}},
            {"protected": protected, "header": {"kid": "b"}},
        ]

        def load_key(header, payload):
            self.assertEqual(payload, b"hello")
            kid = header.get("kid")
            if kid == "a":
                return "secret-a"
            return "secret-b"

        jws = JsonWebSignature()
        s = jws.serialize(header, b"hello", load_key)
        self.assertIsInstance(s, dict)
        self.assertIn("signatures", s)

        data = jws.deserialize(json.dumps(s), load_key)
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header[0]["alg"], "HS256")
        self.assertNotIn("signature", data)

    def test_serialize_json_empty_payload(self):
        jws = JsonWebSignature()
        protected = {"alg": "HS256"}
        header = {"protected": protected, "header": {"kid": "a"}}
        s = jws.serialize_json(header, b"", "secret")
        data = jws.deserialize_json(s, "secret")
        self.assertEqual(data["payload"], b"")

    def test_fail_deserialize_json(self):
        jws = JsonWebSignature()
        self.assertRaises(errors.DecodeError, jws.deserialize_json, None, "")
        self.assertRaises(errors.DecodeError, jws.deserialize_json, "[]", "")
        self.assertRaises(errors.DecodeError, jws.deserialize_json, "{}", "")

        # missing protected
        s = json.dumps({"payload": "YQ"})
        self.assertRaises(errors.DecodeError, jws.deserialize_json, s, "")

        # missing signature
        s = json.dumps({"payload": "YQ", "protected": "YQ"})
        self.assertRaises(errors.DecodeError, jws.deserialize_json, s, "")

    def test_validate_header(self):
        jws = JsonWebSignature(private_headers=[])
        protected = {"alg": "HS256", "invalid": "k"}
        header = {"protected": protected, "header": {"kid": "a"}}
        self.assertRaises(
            errors.InvalidHeaderParameterNameError,
            jws.serialize,
            header,
            b"hello",
            "secret",
        )
        jws = JsonWebSignature(private_headers=["invalid"])
        s = jws.serialize(header, b"hello", "secret")
        self.assertIsInstance(s, dict)

        jws = JsonWebSignature()
        s = jws.serialize(header, b"hello", "secret")
        self.assertIsInstance(s, dict)

    def test_ES512_alg(self):
        jws = JsonWebSignature()
        private_key = read_file_path("secp521r1-private.json")
        public_key = read_file_path("secp521r1-public.json")
        self.assertRaises(
            ValueError, jws.serialize, {"alg": "ES256"}, "hello", private_key
        )
        s = jws.serialize({"alg": "ES512"}, "hello", private_key)
        data = jws.deserialize(s, public_key)
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header["alg"], "ES512")

    def test_ES256K_alg(self):
        jws = JsonWebSignature(algorithms=["ES256K"])
        private_key = read_file_path("secp256k1-private.pem")
        public_key = read_file_path("secp256k1-pub.pem")
        s = jws.serialize({"alg": "ES256K"}, "hello", private_key)
        data = jws.deserialize(s, public_key)
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header["alg"], "ES256K")



import datetime
import unittest

from authlib.jose import JsonWebKey
from authlib.jose import JsonWebToken
from authlib.jose import JWTClaims
from authlib.jose import errors
from authlib.jose import jwt
from authlib.jose.errors import UnsupportedAlgorithmError
from tests.util import read_file_path


class JWTTest(unittest.TestCase):
    def test_init_algorithms(self):
        _jwt = JsonWebToken(["RS256"])
        self.assertRaises(
            UnsupportedAlgorithmError, _jwt.encode, {"alg": "HS256"}, {}, "k"
        )

        _jwt = JsonWebToken("RS256")
        self.assertRaises(
            UnsupportedAlgorithmError, _jwt.encode, {"alg": "HS256"}, {}, "k"
        )

    def test_encode_sensitive_data(self):
        # check=False won't raise error
        jwt.encode({"alg": "HS256"}, {"password": ""}, "k", check=False)
        self.assertRaises(
            errors.InsecureClaimError,
            jwt.encode,
            {"alg": "HS256"},
            {"password": ""},
            "k",
        )
        self.assertRaises(
            errors.InsecureClaimError,
            jwt.encode,
            {"alg": "HS256"},
            {"text": "4242424242424242"},
            "k",
        )

    def test_encode_datetime(self):
        now = datetime.datetime.utcnow()
        id_token = jwt.encode({"alg": "HS256"}, {"exp": now}, "k")
        claims = jwt.decode(id_token, "k")
        self.assertIsInstance(claims.exp, int)

    def test_validate_essential_claims(self):
        id_token = jwt.encode({"alg": "HS256"}, {"iss": "foo"}, "k")
        claims_options = {"iss": {"essential": True, "values": ["foo"]}}
        claims = jwt.decode(id_token, "k", claims_options=claims_options)
        claims.validate()

        claims.options = {"sub": {"essential": True}}
        self.assertRaises(errors.MissingClaimError, claims.validate)

    def test_attribute_error(self):
        claims = JWTClaims({"iss": "foo"}, {"alg": "HS256"})
        self.assertRaises(AttributeError, lambda: claims.invalid)

    def test_invalid_values(self):
        id_token = jwt.encode({"alg": "HS256"}, {"iss": "foo"}, "k")
        claims_options = {"iss": {"values": ["bar"]}}
        claims = jwt.decode(id_token, "k", claims_options=claims_options)
        self.assertRaises(
            errors.InvalidClaimError,
            claims.validate,
        )
        claims.options = {"iss": {"value": "bar"}}
        self.assertRaises(
            errors.InvalidClaimError,
            claims.validate,
        )

    def test_validate_expected_issuer_received_None(self):
        id_token = jwt.encode({"alg": "HS256"}, {"iss": None, "sub": None}, "k")
        claims_options = {"iss": {"essential": True, "values": ["foo"]}}
        claims = jwt.decode(id_token, "k", claims_options=claims_options)
        self.assertRaises(errors.InvalidClaimError, claims.validate)

    def test_validate_aud(self):
        id_token = jwt.encode({"alg": "HS256"}, {"aud": "foo"}, "k")
        claims_options = {"aud": {"essential": True, "value": "foo"}}
        claims = jwt.decode(id_token, "k", claims_options=claims_options)
        claims.validate()

        claims.options = {"aud": {"values": ["bar"]}}
        self.assertRaises(errors.InvalidClaimError, claims.validate)

        id_token = jwt.encode({"alg": "HS256"}, {"aud": ["foo", "bar"]}, "k")
        claims = jwt.decode(id_token, "k", claims_options=claims_options)
        claims.validate()
        # no validate
        claims.options = {"aud": {"values": []}}
        claims.validate()

    def test_validate_exp(self):
        id_token = jwt.encode({"alg": "HS256"}, {"exp": "invalid"}, "k")
        claims = jwt.decode(id_token, "k")
        self.assertRaises(errors.InvalidClaimError, claims.validate)

        id_token = jwt.encode({"alg": "HS256"}, {"exp": 1234}, "k")
        claims = jwt.decode(id_token, "k")
        self.assertRaises(errors.ExpiredTokenError, claims.validate)

    def test_validate_nbf(self):
        id_token = jwt.encode({"alg": "HS256"}, {"nbf": "invalid"}, "k")
        claims = jwt.decode(id_token, "k")
        self.assertRaises(errors.InvalidClaimError, claims.validate)

        id_token = jwt.encode({"alg": "HS256"}, {"nbf": 1234}, "k")
        claims = jwt.decode(id_token, "k")
        claims.validate()

        id_token = jwt.encode({"alg": "HS256"}, {"nbf": 1234}, "k")
        claims = jwt.decode(id_token, "k")
        self.assertRaises(errors.InvalidTokenError, claims.validate, 123)

    def test_validate_iat_issued_in_future(self):
        in_future = datetime.datetime.utcnow() + datetime.timedelta(seconds=10)
        id_token = jwt.encode({"alg": "HS256"}, {"iat": in_future}, "k")
        claims = jwt.decode(id_token, "k")
        with self.assertRaises(errors.InvalidTokenError) as error_ctx:
            claims.validate()
        self.assertEqual(
            str(error_ctx.exception),
            "invalid_token: The token is not valid as it was issued in the future",
        )

    def test_validate_iat_issued_in_future_with_insufficient_leeway(self):
        in_future = datetime.datetime.utcnow() + datetime.timedelta(seconds=10)
        id_token = jwt.encode({"alg": "HS256"}, {"iat": in_future}, "k")
        claims = jwt.decode(id_token, "k")
        with self.assertRaises(errors.InvalidTokenError) as error_ctx:
            claims.validate(leeway=5)
        self.assertEqual(
            str(error_ctx.exception),
            "invalid_token: The token is not valid as it was issued in the future",
        )

    def test_validate_iat_issued_in_future_with_sufficient_leeway(self):
        in_future = datetime.datetime.utcnow() + datetime.timedelta(seconds=10)
        id_token = jwt.encode({"alg": "HS256"}, {"iat": in_future}, "k")
        claims = jwt.decode(id_token, "k")
        claims.validate(leeway=20)

    def test_validate_iat_issued_in_past(self):
        in_future = datetime.datetime.utcnow() - datetime.timedelta(seconds=10)
        id_token = jwt.encode({"alg": "HS256"}, {"iat": in_future}, "k")
        claims = jwt.decode(id_token, "k")
        claims.validate()

    def test_validate_iat(self):
        id_token = jwt.encode({"alg": "HS256"}, {"iat": "invalid"}, "k")
        claims = jwt.decode(id_token, "k")
        self.assertRaises(errors.InvalidClaimError, claims.validate)

    def test_validate_jti(self):
        id_token = jwt.encode({"alg": "HS256"}, {"jti": "bar"}, "k")
        claims_options = {"jti": {"validate": lambda c, o: o == "foo"}}
        claims = jwt.decode(id_token, "k", claims_options=claims_options)
        self.assertRaises(errors.InvalidClaimError, claims.validate)

    def test_validate_custom(self):
        id_token = jwt.encode({"alg": "HS256"}, {"custom": "foo"}, "k")
        claims_options = {"custom": {"validate": lambda c, o: o == "bar"}}
        claims = jwt.decode(id_token, "k", claims_options=claims_options)
        self.assertRaises(errors.InvalidClaimError, claims.validate)

    def test_use_jws(self):
        payload = {"name": "hi"}
        private_key = read_file_path("rsa_private.pem")
        pub_key = read_file_path("rsa_public.pem")
        data = jwt.encode({"alg": "RS256"}, payload, private_key)
        self.assertEqual(data.count(b"."), 2)

        claims = jwt.decode(data, pub_key)
        self.assertEqual(claims["name"], "hi")

    def test_use_jwe(self):
        payload = {"name": "hi"}
        private_key = read_file_path("rsa_private.pem")
        pub_key = read_file_path("rsa_public.pem")
        _jwt = JsonWebToken(["RSA-OAEP", "A256GCM"])
        data = _jwt.encode({"alg": "RSA-OAEP", "enc": "A256GCM"}, payload, pub_key)
        self.assertEqual(data.count(b"."), 4)

        claims = _jwt.decode(data, private_key)
        self.assertEqual(claims["name"], "hi")

    def test_use_jwks(self):
        header = {"alg": "RS256", "kid": "abc"}
        payload = {"name": "hi"}
        private_key = read_file_path("jwks_private.json")
        pub_key = read_file_path("jwks_public.json")
        data = jwt.encode(header, payload, private_key)
        self.assertEqual(data.count(b"."), 2)
        claims = jwt.decode(data, pub_key)
        self.assertEqual(claims["name"], "hi")

    def test_use_jwks_single_kid(self):
        """Test that jwks can be decoded if a kid for decoding is given and encoded data has no kid and only one key is set."""
        header = {"alg": "RS256"}
        payload = {"name": "hi"}
        private_key = read_file_path("jwks_single_private.json")
        pub_key = read_file_path("jwks_single_public.json")
        data = jwt.encode(header, payload, private_key)
        self.assertEqual(data.count(b"."), 2)
        claims = jwt.decode(data, pub_key)
        self.assertEqual(claims["name"], "hi")

    # Added a unit test to showcase my problem.
    # This calls jwt.decode similarly as is done in parse_id_token method of the AsyncOpenIDMixin class when the id token does not contain a kid in the alg header.
    def test_use_jwks_single_kid_keyset(self):
        """Test that jwks can be decoded if a kid for decoding is given and encoded data has no kid and a keyset with one key."""
        header = {"alg": "RS256"}
        payload = {"name": "hi"}
        private_key = read_file_path("jwks_single_private.json")
        pub_key = read_file_path("jwks_single_public.json")
        data = jwt.encode(header, payload, private_key)
        self.assertEqual(data.count(b"."), 2)
        claims = jwt.decode(data, JsonWebKey.import_key_set(pub_key))
        self.assertEqual(claims["name"], "hi")

    def test_with_ec(self):
        payload = {"name": "hi"}
        private_key = read_file_path("secp521r1-private.json")
        pub_key = read_file_path("secp521r1-public.json")
        data = jwt.encode({"alg": "ES512"}, payload, private_key)
        self.assertEqual(data.count(b"."), 2)

        claims = jwt.decode(data, pub_key)
        self.assertEqual(claims["name"], "hi")
